from __future__ import annotations

import logging
from zipfile import ZipFile

from cle.backends.backend import register_backend

from .soot import Soot

log = logging.getLogger(name=__name__)


class Jar(Soot):
    """
    Backend for lifting JARs to Soot.
    """

    is_default = True  # let CLE automatically use this backend

    def __init__(
        self,
        jar_path,
        binary_stream,
        entry_point=None,
        entry_point_params=("java.lang.String[]",),
        jni_libs=None,
        jni_libs_ld_path=None,
        **kwargs,
    ):
        """
        :param jar_path:                Path to JAR.

        The following parameters are optional

        :param entry_point:             Fully qualified name of method that should be used as the entry point.
                                        If not specified, we try to parse it from the manifest.
        :param additional_jars:         Additional JARs.
        :param additional_jar_roots:    Additional JAR roots.
        :param jni_libs:                Name(s) of JNI libs to load (if any).
        :param jni_libs_ld_path:        Path(s) where to find libs defined by param jni_libs.
                                        Note: Directory of the JAR is added by default.
        """

        log.debug("Loading JAR from %s ...", jar_path)

        if not entry_point:
            # try to parse main class from manifest
            self.manifest = self.get_manifest(jar_path)
            main_class = self.manifest.get("Main-Class", None)
            if main_class:
                entry_point = main_class + "." + "main"

        # the actual lifting is done by the Soot superclass
        super().__init__(
            jar_path,
            binary_stream,
            input_format="jar",
            entry_point=entry_point,
            entry_point_params=entry_point_params,
            jni_libs=jni_libs,
            jni_libs_ld_path=jni_libs_ld_path,
            **kwargs,
        )

    @staticmethod
    def is_compatible(stream):
        # check if stream is an archive
        if not Soot.is_zip_archive(stream):
            return False
        # get filelist
        with ZipFile(stream) as jar:
            filelist = jar.namelist()
        # check for manifest and if a least one java class
        # file is available
        if "META-INF/MANIFEST.MF" not in filelist:
            return False
        class_files = [f for f in filelist if f.endswith(".class")]
        if len(class_files) == 0:
            return False
        return True

    def get_manifest(self, binary_path=None):
        """
        Load the MANIFEST.MF file

        :return: A dict of meta info
        :rtype:  dict
        """
        path = binary_path if binary_path else self.binary
        with ZipFile(path) as jar:
            manifest = jar.open("META-INF/MANIFEST.MF", "r")
            data = {}
            for line in manifest.readlines():
                if b":" in line:
                    key, value = line.split(b":")
                    data[key.strip().decode("utf-8")] = value.strip().decode("utf-8")
            return data


register_backend("jar", Jar)



from __future__ import annotations

import logging

from cle.address_translator import AT
from cle.backends.backend import Backend
from cle.backends.region import Segment
from cle.backends.relocation import Relocation
from cle.backends.symbol import Symbol, SymbolType
from cle.errors import CLEError, CLEOperationError
from cle.utils import ALIGN_UP

from .simdata import SimData, lookup
from .simdata.common import PointTo, SimDataSimpleRelocation

log = logging.getLogger(name=__name__)


__all__ = [
    "ExternSegment",
    "TOCRelocation",
    "ExternObject",
    "KernelObject",
    "PointToPrecise",
]


class ExternSegment(Segment):
    def __init__(self, map_size):
        super().__init__(None, 0, None, map_size)

    def addr_to_offset(self, addr):
        raise CLEOperationError(
            "'offset' operations on the extern object are meaningless as it is not mapped from a file"
        )

    def offset_to_addr(self, offset):
        raise CLEOperationError(
            "'offset' operations on the extern object are meaningless as it is not mapped from a file"
        )

    def contains_offset(self, offset):
        return False

    is_readable = True
    is_writable = True
    is_executable = True


class TOCRelocation(Relocation):
    @property
    def value(self):
        return self.resolvedby.rebased_addr


class ExternObject(Backend):
    def __init__(self, loader, map_size=0, tls_size=0):
        super().__init__("cle##externs", None, loader=loader)
        self._next_object = None
        self._delayed_writes = []

        self.next_addr = 0
        self.map_size = map_size
        self.set_arch(loader.main_object.arch)
        self.provides = "extern-address space"
        self.pic = True
        self._import_symbols = {}
        self._warned_data_import = False

        self.tls_data_size = tls_size
        self.tls_next_addr = 0
        self._tls_mapped = False

    def _finalize_tls(self):
        if self._is_mapped or self._tls_mapped:
            raise Exception("programming error")

        if self.tls_data_size != 0:
            self.tls_used = True
            self.tls_data_start = self._allocate(self.tls_data_size, alignment=0x10)
            self.tls_block_size = self.tls_data_size
            self._tls_mapped = True

    def rebase(self, new_base):
        if self._is_mapped:
            return

        if not self._tls_mapped:
            self._finalize_tls()

        backer = bytearray(self.map_size)
        for simdata in self._delayed_writes:
            value = simdata.value()
            start_addr = simdata.relative_addr
            if simdata.type == SymbolType.TYPE_TLS_OBJECT:
                start_addr += self.tls_data_size
            backer[start_addr : start_addr + len(value)] = value

        self.memory.add_backer(0, bytes(backer))
        self.segments.append(ExternSegment(self.map_size))
        super().rebase(new_base)

    def make_extern(
        self, name, size=0, alignment=None, thumb=False, sym_type=SymbolType.TYPE_FUNCTION, point_to=None, libname=None
    ) -> Symbol:
        try:
            return self._symbol_cache[name]
        except KeyError:
            pass

        tls = sym_type == SymbolType.TYPE_TLS_OBJECT
        SymbolCls = Symbol
        if point_to is not None:
            simdata = PointToPrecise
        else:
            simdata = lookup(name, libname)
        if simdata is not None:
            SymbolCls = simdata
            size = simdata.static_size(self)
            if sym_type != simdata.type:
                log.warning("Symbol type mismatch between export request and response for %s. What's going on?", name)

        real_size = max(size, 1)

        if alignment is None:
            alignment = self.arch.bytes

        make_toc = getattr(self.loader.main_object, "is_ppc64_abiv1", False) and sym_type == SymbolType.TYPE_FUNCTION
        toc_symbol = None
        if make_toc:
            # we make two symbols, one for the func and one for the toc
            # the one for the func ends up named with the #func suffix, the toc gets the normal name
            # we return the one for the toc
            toc_symbol = self.make_extern(name, size=0x18, alignment=8, sym_type=SymbolType.TYPE_OBJECT)
            name += "#func"

        if size == 0 and sym_type in (SymbolType.TYPE_NONE, SymbolType.TYPE_OBJECT, SymbolType.TYPE_TLS_OBJECT):
            log.warning(
                "Symbol was allocated without a known size; emulation may fail if it is used non-opaquely: %s", name
            )
            self._warned_data_import = True
            real_size = 8

        local_addr = self._allocate(real_size, alignment=alignment, thumb=thumb, tls=tls)
        if local_addr is None:
            if self._next_object is None:
                # we're at the end of the line. make a new extern object
                # this should only be hit if we're doing this outside a loading pass
                self._make_new_externs(real_size, alignment, tls)
            return self._next_object.make_extern(
                name, size=size, alignment=alignment, sym_type=sym_type, libname=libname
            )

        log.info("Created extern symbol for %s", name)

        new_symbol = SymbolCls(self, name, local_addr, size, sym_type)
        new_symbol.is_export = True
        new_symbol.is_extern = True

        if point_to is not None:
            new_symbol.pointto_name = point_to.name
            new_symbol.pointto_type = point_to.type
            new_symbol.pointto_precise = point_to

        self._symbol_cache[name] = new_symbol
        self.symbols.add(new_symbol)
        self._init_symbol(new_symbol)

        if make_toc:
            # write the pointer to the func into the toc
            # i.e. make a relocation for it
            # then if we're already mapped, apply the relocation manually
            reloc = TOCRelocation(self, toc_symbol, toc_symbol.relative_addr)
            reloc.resolve(new_symbol)
            self.relocs.append(reloc)
            if self._is_mapped:
                reloc.relocate()

            return toc_symbol
        return new_symbol

    def get_pseudo_addr(self, name) -> int:
        if not self._is_mapped:
            raise CLEError("Can't allocate with extern object before it is mapped")

        return self.make_extern(name).rebased_addr

    def allocate(self, size=1, alignment=8, thumb=False, tls=False) -> int:
        if not self._is_mapped:
            raise CLEError("Can't allocate with extern object before it is mapped")

        result = self._allocate(size=size, alignment=alignment, thumb=thumb, tls=tls)
        if result is None:
            if self._next_object is None:
                # we're at the end of the line. make a new extern object
                # this should only be hit if we're doing this outside a loading pass
                self._make_new_externs(size, alignment, tls)
            result = self._next_object.allocate(size=size, alignment=alignment, thumb=thumb, tls=tls)
        return result + (0 if tls else self.mapped_base)

    def _make_new_externs(self, size, alignment, tls):
        self._next_object = ExternObject(
            self.loader,
            map_size=max(size + alignment, 0x8000) if not tls else 0x8000,
            tls_size=max(size + alignment, 0x1000) if tls else 0x1000,
        )
        self._next_object._finalize_tls()
        self.loader._internal_load(self._next_object)

    def _allocate(self, size=1, alignment=8, thumb=False, tls=False):
        if tls:
            start = self.tls_next_addr
            limit = self.tls_data_size
        else:
            start = self.next_addr
            limit = self.map_size

        addr = ALIGN_UP(start, alignment) | thumb
        next_start = addr + size
        if next_start >= limit:
            if self._is_mapped:
                return None
            else:
                if tls:
                    self.tls_data_size += next_start - limit
                else:
                    self.map_size += next_start - limit

        if tls:
            self.tls_next_addr = next_start
            return addr
        else:
            self.next_addr = next_start
            return addr

    @property
    def max_addr(self):
        return AT.from_rva(self.map_size - 1, self).to_mva()

    def make_import(self, name, sym_type):
        if name not in self.imports:
            sym = Symbol(self, name, 0, 0, sym_type)
            sym.is_import = True
            sym.is_extern = True
            # this is kind of tricky... normally if you have an import and an export of the same name in the binary
            # the two symbols are *the same symbol*, usually with a copy relocation. but we don't know ahead of time
            # whether we will have the symbol here in externs, so we will not expose the import symbol to the rest of
            # the world.
            self._import_symbols[name] = sym
            return sym
        else:
            sym = self._import_symbols[name]
            if sym.type != sym_type:
                raise CLEOperationError(
                    "Created the same extern import %s with two different types. Something isn't right!"
                )
            return sym

    def _init_symbol(self, symbol):
        if isinstance(symbol, SimData):
            relocs = symbol.relocations()
            self.relocs.extend(relocs)

            if self._is_mapped:
                # TODO: is this right for tls?
                if symbol.type == SymbolType.TYPE_TLS_OBJECT:
                    self.memory.store(self.tls_block_size, symbol.value())
                else:
                    self.memory.store(symbol.relative_addr, symbol.value())

                for reloc in relocs:
                    reloc.relocate()
            else:
                self._delayed_writes.append(symbol)


class KernelObject(Backend):
    def __init__(self, loader, map_size=0x8000):
        super().__init__("cle##kernel", None, loader=loader)
        self.map_size = map_size
        self.set_arch(loader.main_object.arch)
        self.memory.add_backer(0, bytes(map_size))
        self.provides = "kernel space"
        self.pic = True

    def add_name(self, name, addr):
        self._symbol_cache[name] = Symbol(self, name, AT.from_mva(addr, self).to_rva(), 1, SymbolType.TYPE_FUNCTION)

    @property
    def max_addr(self):
        return AT.from_rva(self.map_size - 1, self).to_mva()


class PointToPrecise(PointTo):
    pointto_precise = None

    def relocations(self):
        return [
            SimDataSimpleRelocation(
                self.owner,
                self.pointto_precise,
                self.relative_addr,
                self.addend,
                preresolved=True,
            )
        ]



from __future__ import annotations

import hashlib
import logging
import os
from io import BufferedReader
from typing import TYPE_CHECKING, Any, BinaryIO

import archinfo
import sortedcontainers

from cle.address_translator import AT
from cle.errors import CLEError, CLEOperationError
from cle.memory import Clemory

from .regions import Regions
from .relocation import Relocation
from .symbol import Symbol

if TYPE_CHECKING:
    from cle.backends import Section, Segment
    from cle.loader import Loader

log = logging.getLogger(name=__name__)


class FunctionHintSource:
    """
    Enums that describe the source of function hints.
    """

    EH_FRAME = 0
    EXTERNAL_EH_FRAME = 1


class FunctionHint:
    """
    Describes a function hint.

    :ivar int addr:     Address of the function.
    :ivar int size:     Size of the function.
    :ivar source:       Source of this hint.
    :vartype source:    int
    """

    __slots__ = ("addr", "size", "source")

    def __init__(self, addr, size, source):
        self.addr = addr
        self.size = size
        self.source = source

    def __repr__(self):
        return f"<FuncHint@{self.addr:#x}, {self.size} bytes>"


class ExceptionHandling:
    """
    Describes an exception handling.

    Exception handlers are usually language-specific. In C++, it is usually implemented as try {} catch {} blocks.

    :ivar int start_addr:               The beginning of the try block.
    :ivar int size:                     Size of the try block.
    :ivar Optional[int] handler_addr:   Address of the exception handler code.
    :ivar type:                         Type of the exception handler. Optional.
    :ivar Optional[int] func_addr:      Address of the function. Optional.
    """

    __slots__ = (
        "start_addr",
        "size",
        "handler_addr",
        "type",
        "func_addr",
    )

    def __init__(self, start_addr, size, handler_addr=None, type_=None, func_addr=None):
        self.start_addr = start_addr
        self.size = size
        self.handler_addr = handler_addr
        self.type = type_
        self.func_addr = func_addr

    def __repr__(self):
        if self.handler_addr is not None:
            return (
                f"<ExceptionHandling@{self.start_addr:#x}-{self.start_addr + self.size:#x}: "
                f"handler@{self.handler_addr:#x}>"
            )
        else:
            return f"<ExceptionHandling@{self.start_addr:#x}-{self.start_addr + self.size:#x}: no handler>"


class Backend:
    """
    Main base class for CLE binary objects.

    An alternate interface to this constructor exists as the static method :meth:`cle.loader.Loader.load_object`

    :ivar binary:           The path to the file this object is loaded from
    :ivar binary_basename:  The basename of the filepath, or a short representation of the stream it was loaded from
    :ivar is_main_bin:      Whether this binary is loaded as the main executable
    :ivar segments:         A listing of all the loaded segments in this file
    :ivar sections:         A listing of all the demarked sections in the file
    :ivar sections_map:     A dict mapping from section name to section
    :ivar imports:          A mapping from symbol name to import relocation
    :ivar resolved_imports: A list of all the import symbols that are successfully resolved
    :ivar relocs:           A list of all the relocations in this binary
    :ivar irelatives:       A list of tuples representing all the irelative relocations that need to be performed. The
                            first item in the tuple is the address of the resolver function, and the second item is the
                            address of where to write the result. The destination address is an RVA.
    :ivar jmprel:           A mapping from symbol name to the address of its jump slot relocation, i.e. its GOT entry.
    :ivar arch:             The architecture of this binary
    :vartype arch:          archinfo.arch.Arch
    :ivar str os:           The operating system this binary is meant to run under
    :ivar int mapped_base:  The base address of this object in virtual memory
    :ivar deps:             A list of names of shared libraries this binary depends on
    :ivar linking:          'dynamic' or 'static'
    :ivar linked_base:      The base address this object requests to be loaded at
    :ivar bool pic:         Whether this object is position-independent
    :ivar bool execstack:   Whether this executable has an executable stack
    :ivar str provides:     The name of the shared library dependancy that this object resolves
    :ivar list symbols:     A list of symbols provided by this object, sorted by address
    :ivar has_memory:       Whether this backend is backed by a Clemory or not. As it stands now, a backend should still
                            define `min_addr` and `max_addr` even if `has_memory` is False.
    """

    is_default = False
    is_outer = False

    def __init__(
        self,
        binary,
        binary_stream,
        loader=None,
        is_main_bin=False,
        entry_point=None,
        arch=None,
        base_addr=None,
        force_rebase=False,
        has_memory=True,
        **kwargs,
    ):
        """
        :param binary:          The path to the binary to load
        :param binary_stream:   The open stream to this binary. The reference to this will be held until you call close.
        :param is_main_bin:     Whether this binary should be loaded as the main executable
        """

        self.load_args: dict[str, Any] = {} | kwargs
        self.set_load_args(
            loader=loader,
            is_main_bin=is_main_bin,
            entry_point=entry_point,
            arch=arch,
            base_addr=base_addr,
            force_rebase=force_rebase,
            has_memory=has_memory,
        )
        self.binary = binary
        self._binary_stream: BufferedReader = binary_stream
        if self.binary is not None:
            self.binary_basename = os.path.basename(self.binary)
        elif hasattr(self._binary_stream, "name"):
            self.binary_basename = os.path.basename(self._binary_stream.name)
        else:
            self.binary_basename = str(self._binary_stream)
        # if the backend unpacks another file, this field will hold the name of the file
        self.unpacked_name: str | None = None

        for k in list(kwargs.keys()):
            if k == "custom_entry_point":
                entry_point = kwargs.pop(k)
            elif k == "custom_arch":
                arch = kwargs.pop(k)
            elif k == "custom_base_addr":
                base_addr = kwargs.pop(k)
            else:
                continue
            log.critical("Deprecation warning: the %s parameter has been renamed to %s", k, k[7:])

        if kwargs != {}:
            log.warning("Unused kwargs for loading binary %s: %s", self.binary, ", ".join(kwargs.keys()))

        self.is_main_bin = is_main_bin
        self.has_memory = has_memory
        self._loader: Loader | None = loader
        self._entry = 0
        self._segments: Regions[Segment] = Regions()  # List of segments
        self._sections: Regions[Section] = Regions()  # List of sections
        self.sections_map = {}  # Mapping from section name to section
        self.symbols = sortedcontainers.SortedKeyList(key=self._get_symbol_relative_addr)
        self.imports: dict[str, Relocation] = {}
        self.resolved_imports = []
        self.relocs: list[Relocation] = []
        self.irelatives = []  # list of tuples (resolver, destination), dest w/o rebase
        self.jmprel = {}
        self._arch: archinfo.Arch | None = None
        self.os = None  # Let other stuff override this
        self.compiler = None, None  # compiler name, version
        self._symbol_cache = {}
        # a list of directories to search for libraries specified by the object
        self.extra_load_path = []
        # attributes to enable SimProcedure guessing
        self.guess_simprocs = False
        self.guess_simprocs_hint = None
        # if we want one of our children to be the main object of the loader, we set this field to the child, and the
        # loader will pick it up
        self.force_main_object = None

        # checksums
        self.md5 = None
        self.sha256 = None

        self.mapped_base_symbolic = 0
        # These are set by cle, and should not be overriden manually
        self.mapped_base = self.linked_base = 0  # not to be set manually - used by CLE

        self.deps = []  # Needed shared objects (libraries dependencies)
        self.child_objects: list[Backend] = []  # any objects loaded directly out of this
        self.parent_object = None
        self.linking = None  # Dynamic or static linking
        self.pic = force_rebase
        self.execstack = False
        self.aslr = False

        # tls info set by backend to communicate with thread manager
        self.tls_used = False
        self.tls_block_size = None
        self.tls_data_size = None
        self.tls_data_start = None
        # tls info set by thread manager
        self.tls_module_id = None
        # self.tls_block_offset = None  # this is an ELF-only attribute

        # exception handling
        # they should be rebased when .rebase() is called
        self.exception_handlings: list[ExceptionHandling] = []

        # Hints
        # they should be rebased when .rebase() is called
        self.function_hints: list[FunctionHint] = []

        # line number mapping
        self.addr_to_line = {}

        # Custom options
        self._custom_entry_point = entry_point
        self._custom_base_addr = base_addr
        self.provides = os.path.basename(self.binary) if self.binary is not None else None

        self.memory: Clemory

        # should be set inside `cle.Loader.add_object`
        self._is_mapped = False
        # cached max_addr
        self._max_addr = None
        # cached last section
        self._last_section = None
        # cached last segment
        self._last_segment = None

        self.cached_content: bytes | None = None

        if arch is None:
            pass
        elif isinstance(arch, str):
            self.set_arch(archinfo.arch_from_id(arch))
        elif isinstance(arch, archinfo.Arch):
            self.set_arch(arch)
        elif isinstance(arch, type) and issubclass(arch, archinfo.Arch):
            self.set_arch(arch())  # type: ignore
        else:
            raise CLEError(f"Bad parameter: arch={arch}")

        self._cache_content()
        self._checksum()

    @property
    def arch(self) -> archinfo.Arch:
        result = self._arch
        if result is None:
            raise ValueError("No arch is assigned yet")
        return result

    @property
    def loader(self) -> Loader:
        result = self._loader
        if result is None:
            raise ValueError("Backend does not have a loader associated")
        return result

    def close(self) -> None:
        del self._binary_stream

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} Object {self.binary_basename}, maps [{self.min_addr:#x}:{self.max_addr:#x}]>"
        )

    def set_arch(self, arch):
        self._arch = arch
        self.memory = Clemory(arch)  # Private virtual address space, without relocations

    def set_load_args(self, **kwargs) -> None:
        self.load_args |= kwargs

    @property
    def image_base_delta(self):
        return self.mapped_base - self.linked_base

    @property
    def entry(self):
        if self._custom_entry_point is not None:
            return AT.from_lva(self._custom_entry_point, self).to_mva()
        return AT.from_lva(self._entry, self).to_mva()

    @property
    def segments(self) -> Regions[Segment]:
        return self._segments

    @segments.setter
    def segments(self, v: Regions[Segment] | list[Segment]):
        if isinstance(v, list):
            self._segments = Regions(lst=v)
        elif isinstance(v, Regions):
            self._segments = v
        else:
            raise ValueError(f"Unsupported type {type(v)} set as sections.")

    @property
    def sections(self) -> Regions[Section]:
        return self._sections

    @sections.setter
    def sections(self, v: Regions[Section] | list[Section]):
        if isinstance(v, list):
            self._sections = Regions(lst=v)
        elif isinstance(v, Regions):
            self._sections = v
        else:
            raise ValueError(f"Unsupported type {type(v)} set as sections.")

    @property
    def symbols_by_addr(self):
        log.critical(
            "Deprecation warning: symbols_by_addr is deprecated - use loader.find_symbol() for lookup "
            "and .symbols for enumeration"
        )
        return {s.rebased_addr: s for s in self.symbols}

    def rebase(self, new_base):
        """
        Rebase backend's regions to the new base where they were mapped by the loader
        """
        if self._is_mapped:
            # we could rebase an object twice if we really wanted... no need though, right?
            raise CLEOperationError(f"Image already rebased from {self.linked_base:#x} to {self.mapped_base:#x}")

        self.mapped_base = new_base

        if self.sections:
            self.sections._rebase(self.image_base_delta)
        if self.segments and self.sections is not self.segments:
            self.segments._rebase(self.image_base_delta)

        for handling in self.exception_handlings:
            if handling.func_addr is not None:
                handling.func_addr += self.image_base_delta
            if handling.handler_addr is not None:
                handling.handler_addr += self.image_base_delta
            handling.start_addr += self.image_base_delta

        for hint in self.function_hints:
            hint.addr = hint.addr + self.image_base_delta

    def relocate(self):
        """
        Apply all resolved relocations to memory.

        The meaning of "resolved relocations" is somewhat subtle - there is a linking step which attempts to resolve
        each relocation, currently only present in the main internal loading function since the calculation of which
        objects should be available
        """
        for reloc in self.relocs:
            if reloc.resolved:
                reloc.relocate()

    def contains_addr(self, addr):
        """
        Is `addr` in one of the binary's segments/sections we have loaded? (i.e. is it mapped into memory ?)
        """
        return self.find_loadable_containing(addr) is not None

    def find_loadable_containing(self, addr):
        lookup = self.find_segment_containing if self.segments else self.find_section_containing
        return lookup(addr)

    def find_segment_containing(self, addr: int) -> Segment | None:
        """
        Returns the segment that contains `addr`, or ``None``.
        """
        if self._last_segment is not None and self._last_segment.contains_addr(addr):
            return self._last_segment

        r = self.segments.find_region_containing(addr)
        if r is not None:
            self._last_segment = r
        return r

    def find_section_containing(self, addr: int) -> Section | None:
        """
        Returns the section that contains `addr` or ``None``.
        """
        if self._last_section is not None and self._last_section.contains_addr(addr):
            return self._last_section

        r = self.sections.find_region_containing(addr)
        if r is not None:
            self._last_section = r
        return r

    def addr_to_offset(self, addr: int) -> int | None:
        loadable = self.find_loadable_containing(addr)
        if loadable is not None:
            return loadable.addr_to_offset(addr)
        else:
            return None

    def offset_to_addr(self, offset: int) -> int | None:
        if self.segments:
            for s in self.segments:
                if s.contains_offset(offset):
                    return s.offset_to_addr(offset)
        else:
            for s in self.sections:
                if s.contains_offset(offset):
                    return s.offset_to_addr(offset)
        return None

    @property
    def min_addr(self) -> int:
        """
        This returns the lowest virtual address contained in any loaded segment of the binary.
        """
        # Loader maps the object at chosen mapped base anyway and independently of the internal structure
        return self.mapped_base

    @property
    def max_addr(self) -> int:
        """
        This returns the highest virtual address contained in any loaded segment of the binary.
        """

        if self._max_addr is None:
            out = self.mapped_base
            if self.segments or self.sections:
                out = max(x.max_addr for x in (self.segments or self.sections))
            self._max_addr = out - self.mapped_base
        return self._max_addr + self.mapped_base

    @property
    def initializers(self) -> list[int]:  # pylint: disable=no-self-use
        """
        Stub function. Should be overridden by backends that can provide initializer functions that ought to be run
        before execution reaches the entry point. Addresses should be rebased.
        """
        return []

    @property
    def finalizers(self) -> list[int]:  # pylint: disable=no-self-use
        """
        Stub function. Like initializers, but with finalizers.
        """
        return []

    @property
    def threads(self) -> list:  # pylint: disable=no-self-use
        """
        If this backend represents a dump of a running program, it may contain one or more thread contexts, i.e.
        register files. This property should contain a list of names for these threads, which should be unique.
        """
        return []

    def thread_registers(self, thread=None) -> dict[str, Any]:  # pylint: disable=no-self-use,unused-argument
        """
        If this backend represents a dump of a running program, it may contain one or more thread contexts, i.e.
        register files. This method should return the register file for a given thread (as named in ``Backend.threads``)
        as a dict mapping register names (as seen in archinfo) to numbers. If the thread is not specified, it should
        return the context for a "default" thread. If there are no threads, it should return an empty dict.
        """
        return {}

    def initial_register_values(self):
        """
        Deprecated
        """
        log.critical(
            "Deprecation warning: initial_register_values is deprecated - " "use backend.thread_registers() instead"
        )
        return self.thread_registers().items()

    def get_symbol(self, name: str) -> Symbol | None:  # pylint: disable=no-self-use,unused-argument
        """
        Stub function. Implement to find the symbol with name `name`.
        """
        if name in self._symbol_cache:
            return self._symbol_cache[name]
        for sym in self.symbols:
            if sym.name == name:
                self._symbol_cache[name] = sym
                return sym
        return None

    @staticmethod
    def extract_soname(path) -> str | None:  # pylint: disable=unused-argument
        """
        Extracts the shared object identifier from the path, or returns None if it cannot.
        """
        return None

    @classmethod
    def is_compatible(cls, stream) -> bool:  # pylint:disable=unused-argument
        """
        Determine quickly whether this backend can load an object from this stream
        """
        return False

    @classmethod
    def check_compatibility(cls, spec, obj) -> bool:  # pylint: disable=unused-argument
        """
        Performs a minimal static load of ``spec`` and returns whether it's compatible with other_obj
        """
        return False

    @classmethod
    def check_magic_compatibility(cls, stream: BinaryIO) -> bool:  # pylint: disable=unused-argument
        """
        Check if a stream of bytes contains the same magic number as the main object
        """
        return False

    @staticmethod
    def _get_symbol_relative_addr(value):
        return value.relative_addr

    def _cache_content(self):
        """
        Cache the raw content of this object.
        """
        if self._binary_stream is not None:
            self._binary_stream.seek(0)
            data = self._binary_stream.read()
            self._binary_stream.seek(0)
            self.cached_content = data

    def _checksum(self):
        """
        Calculate MD5 and SHA256 checksum for the binary.
        """

        if self._binary_stream is not None:
            self._binary_stream.seek(0)
            data = self._binary_stream.read()
            self._binary_stream.seek(0)
            self.md5 = hashlib.md5(data).digest()
            self.sha256 = hashlib.sha256(data).digest()

    def __getstate__(self):
        state = self.__dict__.copy()
        state["symbols"] = list(state["symbols"])
        return state

    def __setstate__(self, state):
        state["symbols"] = sortedcontainers.SortedKeyList(state["symbols"], key=self._get_symbol_relative_addr)
        self.__dict__.update(state)
        for sym in self.symbols:
            sym.owner = self

    def __contains__(self, thing: int) -> bool:
        """
        This serves two purposes:
        1. It's slightly more convenient than writing self.min_addr <= thing < self.max_addr yourself
        2. If a Backend implements some form of __getitem__ that always returns False for an integer, running
        `0 in backend` will run into an infinite loop. This prevents that, by just defining sensible semantics for `in`

        This could also be extended to other types, in the future, if it makes sense.
        """

        if isinstance(thing, int):
            return self.min_addr <= thing < self.max_addr
        raise ValueError(f"Unsupported type {type(thing)} for containment check")


ALL_BACKENDS: dict[str, type[Backend]] = {}


def register_backend(name, cls):
    ALL_BACKENDS.update({name: cls})



# pylint:disable=bad-builtin
from __future__ import annotations

import copy
import logging
import os
import pathlib
import xml.etree.ElementTree
from collections import OrderedDict, defaultdict

import archinfo
import elftools
from elftools.common.exceptions import DWARFError, ELFError, ELFParseError
from elftools.dwarf import callframe
from elftools.dwarf.descriptions import describe_attr_value, describe_form_class
from elftools.dwarf.die import DIE
from elftools.dwarf.dwarf_expr import DWARFExprParser
from elftools.dwarf.dwarfinfo import DWARFInfo
from elftools.elf import dynamic, elffile, enums, sections
from sortedcontainers import SortedDict

from cle.address_translator import AT
from cle.backends.backend import ExceptionHandling, FunctionHint, FunctionHintSource, register_backend
from cle.errors import CLECompatibilityError, CLEError, CLEInvalidBinaryError
from cle.patched_stream import PatchedStream
from cle.utils import ALIGN_DOWN, ALIGN_UP, get_mmaped_data, stream_or_path

from .compilation_unit import CompilationUnit
from .hashtable import ELFHashTable, GNUHashTable
from .lsda import LSDAExceptionTable
from .metaelf import MetaELF, maybedecode
from .regions import ELFSection, ELFSegment
from .relocation import get_relocation
from .relocation.generic import MipsGlobalReloc, MipsLocalReloc
from .subprogram import LexicalBlock, Subprogram
from .symbol import ELFSymbol, Symbol, SymbolType
from .variable import Variable
from .variable_type import VariableType

try:
    import pypcode
except ImportError:
    pypcode = None


log = logging.getLogger(name=__name__)

__all__ = ("ELFSymbol", "ELF")


# map 'e_machine' ELF header values (represented as `short int`s) to human-readable format (string)
# There are mappings missing currently in `elftools`, so we provide them ourselves
additional_e_machine_mappings: dict[int, str] = {
    247: "Linux BPF - in-kernel virtual machine",
    252: "C-SKY",
    0x5441: "Fujitsu FR-V",
    0x9026: "DEC Alpha",
}


class ELF(MetaELF):
    """
    The main loader class for statically loading ELF executables. Uses the pyreadelf library where useful.

    Useful backend options:

    - ``debug_symbols``: Provides the path to a separate file which contains the binary's debug symbols
    - ``discard_section_headers``: Do not parse section headers. Use this if they are corrupted or malicious.
    - ``discard_program_headers``: Do not parse program headers. Use this if the binary is for a platform whose ELF
            loader only looks at section headers, but whose toolchain generates program headers anyway.
    """

    is_default = True  # Tell CLE to automatically consider using the ELF backend

    def __init__(
        self,
        *args,
        addend=None,
        debug_symbols=None,
        discard_section_headers=False,
        discard_program_headers=False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.set_load_args(
            addend=addend,
            debug_symbols=debug_symbols,
            discard_section_headers=discard_section_headers,
            discard_program_headers=discard_program_headers,
        )
        patch_undo = []
        try:
            self._reader = elffile.ELFFile(self._binary_stream)
            list(self._reader.iter_sections())
        except Exception as e:  # pylint: disable=broad-except
            self._binary_stream.seek(4)
            ty = self._binary_stream.read(1)
            if ty not in (b"\1", b"\2"):
                raise CLECompatibilityError from e

            if ty == b"\1":
                patch_data = [(0x20, b"\0" * 4), (0x2E, b"\0" * 6)]
            else:
                patch_data = [(0x28, b"\0" * 8), (0x3A, b"\0" * 6)]

            for offset, patch in patch_data:
                self._binary_stream.seek(offset)
                patch_undo.append((offset, self._binary_stream.read(len(patch))))

            self._binary_stream = PatchedStream(self._binary_stream, patch_data)
            log.error("PyReadELF couldn't load this file. Trying again without section headers...")

            try:
                self._reader = elffile.ELFFile(self._binary_stream)
            except Exception as e1:  # pylint: disable=broad-except
                raise CLECompatibilityError from e1

        # Get an appropriate archinfo.Arch for this binary, unless the user specified one
        if self._arch is None:
            self.set_arch(self.extract_arch(self._reader))
        else:
            try:
                other_arch = self.extract_arch(self._reader)
                if other_arch != self.arch:
                    log.warning("User specified %s but autodetected %s. Proceed with caution.", self.arch, other_arch)
            except archinfo.ArchNotFound:
                pass

        self._addend = addend

        # initializers and finalizers
        self._inits_extracted = False
        self._preinit_arr = []
        self._init_func = None
        self._init_arr = []
        self._fini_func = None
        self._fini_arr = []

        # symbols
        self._nullsymbol = Symbol(self, "", 0, 0, SymbolType.TYPE_NONE)
        self._nullsymbol.is_static = True
        self._symbol_cache = {}
        self._symbols_by_name = {}
        self._desperate_for_symbols = False
        self.imports = {}
        self.resolved_imports = []
        self.hashtable = None
        self._versions = {0: "*local*", 1: "*global*"}
        self._vertable = None

        # relocations
        self.relocs = []
        self.jmprel = OrderedDict()
        self.rela_type = None
        self.__parsed_reloc_tables = set()

        # DWARF data
        self.has_dwarf_info = bool(self._reader.has_dwarf_info())
        self.build_id = None
        self.addr_to_line: SortedDict[int, set[tuple[int, int]]] = SortedDict()
        self.variables: list[Variable] | None = None
        self.compilation_units: list[CompilationUnit] | None = None

        # misc
        self._entry = self._reader.header.e_entry
        self.is_relocatable = self._reader.header.e_type == "ET_REL"
        self.pic = self.pic or self._reader.header.e_type in ("ET_REL", "ET_DYN")
        self.tls_block_offset = None  # this is an ELF-only attribute
        self._dynamic = {}
        self.deps = []

        # The linked image base should be evaluated before registering any segment or section due to
        # the fact that elffile, used by those methods, is working only with un-based virtual addresses, but Clemories
        # themselves are organized as a tree where each node backer internally uses relative addressing
        seg_addrs = (
            ALIGN_DOWN(x["p_vaddr"], self.loader.page_size)
            for x in self._reader.iter_segments()
            if x.header.p_type == "PT_LOAD" and x.header.p_memsz > 0
        )
        self.mapped_base = self.linked_base = 0
        try:
            self.mapped_base = self.linked_base = min(seg_addrs)
        except ValueError:
            log.info("no PT_LOAD segments identified")

        if not discard_program_headers:
            self.__register_segments()
        if not discard_section_headers:
            self.__register_sections()

        if not self.symbols:
            self._desperate_for_symbols = True
            self.symbols.update(self._symbol_cache.values())

        if self.has_dwarf_info and self.loader._load_debug_info:
            # load DWARF information
            try:
                dwarf = self._reader.get_dwarf_info()
            except ELFError:
                log.warning(
                    "An exception occurred in pyelftools when loading the DWARF information for %s. "
                    "Marking DWARF as not available for this binary.",
                    self.binary_basename,
                    exc_info=True,
                )
                dwarf = None
                self.has_dwarf_info = False

            if dwarf:
                # Load DIEs
                self._load_dies(dwarf)
                # Load function hints and exception handling artifacts
                if dwarf.has_EH_CFI():
                    self._load_function_hints_from_fde(dwarf, FunctionHintSource.EH_FRAME)
                    self._load_exception_handling(dwarf)
                    self._load_line_info(dwarf)

        if debug_symbols:
            self.__process_debug_file(debug_symbols)
        elif self.loader._load_debug_info:
            if self.build_id:
                debug_filename = f"/usr/lib/debug/.build-id/{self.build_id[:2]}/{self.build_id[2:]}.debug"
                if os.path.isfile(debug_filename):
                    self.__process_debug_file(debug_filename)
            if self.binary:
                debug_filename = os.path.join("/usr/lib/debug", os.path.realpath(self.binary))
                if os.path.isfile(debug_filename):
                    self.__process_debug_file(debug_filename)

        # call the methods defined by MetaELF
        self._ppc64_abiv1_entry_fix()
        self._load_plt()

        # hack: set guess_simprocs = True for object files
        if self.is_relocatable and self.imports and not self._dynamic:
            self.guess_simprocs = True

        for offset, patch in patch_undo:
            self.memory.store(AT.from_lva(self.min_addr + offset, self).to_rva(), patch)

    #
    # Properties and Public Methods
    #

    def close(self):
        super().close()
        del self._reader

    @classmethod
    def check_compatibility(cls, spec, obj):
        with stream_or_path(spec) as stream:
            try:
                return cls.extract_arch(elffile.ELFFile(stream)) == obj.arch
            except archinfo.ArchNotFound:
                return False

    @classmethod
    def check_magic_compatibility(cls, stream):
        stream.seek(0)
        identstring = stream.read(0x10)
        stream.seek(0)
        return identstring.startswith(b"\x7fELF")

    @staticmethod
    def is_compatible(stream):
        stream.seek(0)
        identstring = stream.read(0x1000)
        stream.seek(0)
        if identstring.startswith(b"\x7fELF"):
            if elftools.elf.elffile.ELFFile(stream).header["e_type"] == "ET_CORE":
                return False
            return True
        return False

    @staticmethod
    def _extract_arm_attrs(reader):
        # EDG says: Did you know ARM puts a whole special section in their elves to hold all their ABI junk?
        # As ARM architectures start to diverge a bunch, we actually need this data to pick out
        # what to do.  Particularly for us, it's whether we're doing Cortex-M or not, as our other ARM support
        # is generic

        # TODO: There's a whole pile of useful stuff in here. We should use it some day.
        attrs_sec = reader.get_section_by_name(".ARM.attributes")
        if not attrs_sec:
            return None  # No attrs here!
        attrs_sub_sec = None
        for subsec in attrs_sec.subsections:
            if isinstance(subsec, elftools.elf.sections.ARMAttributesSubsection):
                attrs_sub_sec = subsec
                break
        if not attrs_sub_sec:
            return None  # None here either
        attrs_sub_sub_sec = None
        for subsubsec in attrs_sub_sec.subsubsections:
            if isinstance(subsubsec, elftools.elf.sections.ARMAttributesSubsubsection):
                attrs_sub_sub_sec = subsubsec
                break
        if not attrs_sub_sub_sec:
            return None  # None here either
        # Ok, now we can finally look at the goods
        atts = {}
        for attobj in attrs_sub_sub_sec.attributes:
            atts[attobj.tag] = attobj.value
        return atts

    @staticmethod
    def extract_arch(reader):
        e_machine_header_val = reader["e_machine"]
        arch_str = None
        if isinstance(e_machine_header_val, str):
            arch_str = e_machine_header_val
        elif isinstance(e_machine_header_val, int):
            if e_machine_header_val in additional_e_machine_mappings:
                arch_str = additional_e_machine_mappings[e_machine_header_val]
            else:
                raise CLECompatibilityError(
                    f"The `e_machine` header value of the ELF file is not known: {e_machine_header_val}"
                )
        else:
            assert False  # assumption that `reader['e_machine'] returns `str` or `int` is violated

        if "ARM" in arch_str:
            # Check the ARM attributes, if they exist
            arm_attrs = ELF._extract_arm_attrs(reader)
            if arm_attrs and "TAG_CPU_NAME" in arm_attrs:
                if arm_attrs["TAG_CPU_NAME"].endswith("-M") or "Cortex-M" in arm_attrs["TAG_CPU_NAME"]:
                    return archinfo.ArchARMCortexM("Iend_LE")
            if reader.header.e_flags & 0x200:
                return archinfo.ArchARMEL("Iend_LE" if reader.little_endian else "Iend_BE")
            elif reader.header.e_flags & 0x400:
                return archinfo.ArchARMHF("Iend_LE" if reader.little_endian else "Iend_BE")

        try:
            return archinfo.arch_from_id(arch_str, "le" if reader.little_endian else "be", reader.elfclass)
        except archinfo.ArchNotFound:
            arch = ELF._extract_pcode_arch(reader)
            if arch:
                return arch
            raise

    @staticmethod
    def _extract_pcode_arch(reader):
        if pypcode:
            languages = ELF._get_compatible_pcode_languages(reader)
            if languages:
                return archinfo.ArchPcode(languages[0])
        return None

    @property
    def initializers(self):
        if not self._inits_extracted:
            self._extract_init_fini()
        out = []

        if self.is_main_bin:
            # Preinitializers are ignored in shared objects.
            out.extend(self._preinit_arr)
        else:
            # The init func and the init array in the dynamic section are only run by the dynamic loader in shared
            # objects. In the main binary they are run by libc_csu_init (or libc_start_main in newer glibc)
            if self._init_func is not None:
                out.append(self._init_func)
            out.extend(self._init_arr)

        for i, x in enumerate(out):
            out[i] = AT.from_lva(x, self).to_mva()

        return out

    @property
    def finalizers(self):
        if not self._inits_extracted:
            self._extract_init_fini()
        out = []
        if self._fini_func is not None:
            out.append(self._fini_func)
        out.extend(self._fini_arr)
        for i, x in enumerate(out):
            out[i] = AT.from_lva(x, self).to_mva()
        return out

    @property
    def symbols_by_name(self):
        return self._symbols_by_name.copy()

    def get_symbol(self, symid, symbol_table=None):  # pylint: disable=arguments-differ,arguments-renamed
        """
        Gets a Symbol object for the specified symbol.

        :param symid: Either an index into .dynsym or the name of a symbol.
        """
        version = None
        if isinstance(symid, int):
            if symid == 0:
                # special case the null symbol, this is important for static binaries
                return self._nullsymbol
            if symbol_table is None:
                raise TypeError("Must specify the symbol table to look up symbols by index")
            try:
                re_sym = symbol_table.get_symbol(symid)
            except Exception:  # pylint: disable=broad-except
                log.exception("Error parsing symbol %#08x", symid)
                return None
            cache_key = self._symbol_to_tuple(re_sym)
            cached = self._symbol_cache.get(cache_key, None)
            if cached is not None:
                return cached
            if self.hashtable is not None and symbol_table is self.hashtable.symtab and self._vertable is not None:
                version = self._vertable.get_symbol(symid).entry.ndx
        elif isinstance(symid, str):
            if not symid:
                log.warning("Trying to resolve a symbol by its empty name")
                return None
            cached = self._symbols_by_name.get(symid, None)
            if cached:
                return cached
            if self.hashtable is None:
                return None
            idx, re_sym = self.hashtable.get(symid)
            if re_sym is None:
                return None
            cache_key = self._symbol_to_tuple(re_sym)
            if self._vertable is not None:
                version = self._vertable.get_symbol(idx).entry.ndx
        elif isinstance(symid, sections.Symbol):
            cache_key = self._symbol_to_tuple(symid)
            cached = self._symbol_cache.get(cache_key, None)
            if cached is not None:
                return cached
            re_sym = symid
        else:
            raise CLEError(f"Bad symbol identifier: {symid!r}")

        symbol = ELFSymbol(self, re_sym)
        if version is not None and self._versions is not None:
            version = enums.ENUM_VERSYM.get(version, version) & 0x7FFF
            symbol.version = self._versions[version]
        self._symbol_cache[cache_key] = symbol
        self._cache_symbol_name(symbol)
        return symbol

    def rebase(self, new_base):
        delta = new_base - self.linked_base
        super().rebase(new_base)

        self.addr_to_line = SortedDict((addr + delta, value) for addr, value in self.addr_to_line.items())

    #
    # Private Methods
    #

    @staticmethod
    def _symbol_to_tuple(re_sym):
        """
        Returns a tuple of properties of the given PyELF symbol.
        This is unique enough as a key for both symbol lookup and retrieval.
        """
        entry = re_sym.entry
        return (entry.st_name, entry.st_value, entry.st_size, entry.st_info.bind, entry.st_info.type, entry.st_shndx)

    def _cache_symbol_name(self, symbol):
        name = symbol.name
        if name:
            if self._desperate_for_symbols:
                idx = self.symbols.bisect_key_left(symbol.relative_addr)
                if idx >= len(self.symbols) or self.symbols[idx].name != name:
                    self.symbols.add(symbol)

            if name in self._symbols_by_name:
                old_symbol = self._symbols_by_name[name]
                if not old_symbol.is_weak and symbol.is_weak:
                    return
            self._symbols_by_name[name] = symbol

    def _extract_init_fini(self):
        # Extract the initializers and finalizers
        # the arrays are actually mvas because they are in memory and thus relocated. turn them into lvas.
        if "DT_PREINIT_ARRAY" in self._dynamic and "DT_PREINIT_ARRAYSZ" in self._dynamic:
            arr_start = AT.from_lva(self._dynamic["DT_PREINIT_ARRAY"], self).to_rva()
            arr_end = arr_start + self._dynamic["DT_PREINIT_ARRAYSZ"]
            arr_entsize = self.arch.bytes
            self._preinit_arr = list(map(self.memory.unpack_word, range(arr_start, arr_end, arr_entsize)))
            for i, x in enumerate(self._preinit_arr):
                self._preinit_arr[i] = AT.from_mva(x, self).to_lva()
        if "DT_INIT" in self._dynamic:
            self._init_func = self._dynamic["DT_INIT"]
        if "DT_INIT_ARRAY" in self._dynamic and "DT_INIT_ARRAYSZ" in self._dynamic:
            arr_start = AT.from_lva(self._dynamic["DT_INIT_ARRAY"], self).to_rva()
            arr_end = arr_start + self._dynamic["DT_INIT_ARRAYSZ"]
            arr_entsize = self.arch.bytes
            self._init_arr = list(map(self.memory.unpack_word, range(arr_start, arr_end, arr_entsize)))
        for i, x in enumerate(self._init_arr):
            self._init_arr[i] = AT.from_mva(x, self).to_lva()
        if "DT_FINI" in self._dynamic:
            self._fini_func = self._dynamic["DT_FINI"]
        if "DT_FINI_ARRAY" in self._dynamic and "DT_FINI_ARRAYSZ" in self._dynamic:
            arr_start = AT.from_lva(self._dynamic["DT_FINI_ARRAY"], self).to_rva()
            arr_end = arr_start + self._dynamic["DT_FINI_ARRAYSZ"]
            arr_entsize = self.arch.bytes
            self._fini_arr = list(map(self.memory.unpack_word, range(arr_start, arr_end, arr_entsize)))
            for i, x in enumerate(self._fini_arr):
                self._fini_arr[i] = AT.from_mva(x, self).to_lva()
        self._inits_extracted = True

    def _load_segment(self, seg):
        loaded_segment = ELFSegment(seg)
        self.segments.append(loaded_segment)

        # see https://code.woboq.org/userspace/glibc/elf/dl-load.c.html#1066
        ph = seg.header

        if ph.p_align & (self.loader.page_size - 1) != 0:
            log.error(
                "ELF file %s is loading a segment which is not page-aligned... do you need to change the page size?",
                self.binary,
            )

        if (ph.p_vaddr - ph.p_offset) & (ph.p_align - 1) != 0:
            log.warning(
                "ELF file %s is loading a segment with an inappropriate alignment. It might not work in all contexts.",
                self.binary,
            )
        if ph.p_filesz > ph.p_memsz:
            raise CLEInvalidBinaryError(f"ELF file {self.binary} is loading a segment with an inappropriate allocation")

        mapstart = ALIGN_DOWN(ph.p_vaddr, self.loader.page_size)
        mapend = ALIGN_UP(ph.p_vaddr + ph.p_filesz, self.loader.page_size)

        dataend = ph.p_vaddr + ph.p_filesz
        allocend = ph.p_vaddr + ph.p_memsz

        mapoff = ALIGN_DOWN(ph.p_offset, self.loader.page_size)

        # patch modified addresses into ELFSegment
        loaded_segment.vaddr = mapstart
        loaded_segment.memsize = mapend - mapstart
        loaded_segment.filesize = mapend - mapstart
        loaded_segment.offset = mapoff

        # see https://code.woboq.org/userspace/glibc/elf/dl-map-segments.h.html#88
        data = get_mmaped_data(seg.stream, mapoff, mapend - mapstart, self.loader.page_size)
        if allocend > dataend:
            zero = dataend
            zeropage = (zero + self.loader.page_size - 1) & ~(self.loader.page_size - 1)

            if zeropage > zero:
                data = data[: zero - mapstart].ljust(zeropage - mapstart, b"\0")

            zeroend = ALIGN_UP(allocend, self.loader.page_size)  # mmap maps to the next page boundary
            if zeroend > zeropage:
                data = data.ljust(zeroend - mapstart, b"\0")
            loaded_segment.memsize = zeroend - mapstart
        elif not data:
            log.warning("Segment %s is empty at %#08x!", seg.header.p_type, mapstart)
            return

        self.memory.add_backer(AT.from_lva(mapstart, self).to_rva(), data, overwrite=True)

    def _make_reloc(self, readelf_reloc, symbol, dest_section: ELFSection | None = None):
        addend = readelf_reloc.entry.r_addend if readelf_reloc.is_RELA() else None
        RelocClass = get_relocation(self.arch.name, readelf_reloc.entry.r_info_type)
        if RelocClass is None:
            return None

        address = AT.from_lva(readelf_reloc.entry.r_offset, self).to_rva()
        if dest_section is not None:
            # note to the intrepid explorer: this code was changed as per cle#467
            # address += dest_section.remap_offset
            if self.is_relocatable:
                address += AT.from_mva(dest_section.vaddr, self).to_rva()

        try:
            return RelocClass(self, symbol, address, addend)
        except KeyError:
            log.error("Malformed relocation: access to unmapped %#x", readelf_reloc.entry.r_offset)
            return None

    def _load_function_hints_from_fde(self, dwarf, source):
        """
        Load frame description entries out of the .eh_frame section. These entries include function addresses and can be
        used to improve CFG recovery.

        :param dwarf:   The DWARF info object from pyelftools.
        :return:        None
        """

        try:
            for entry in dwarf.EH_CFI_entries():
                if isinstance(entry, callframe.FDE):
                    self.function_hints.append(
                        FunctionHint(
                            entry.header["initial_location"],
                            entry.header["address_range"],
                            source,
                        )
                    )
        except (DWARFError, ValueError):
            log.warning("An exception occurred in pyelftools when loading FDE information.", exc_info=True)

    def _load_exception_handling(self, dwarf):
        """
        Load exception handling information out of the .eh_frame and .gcc_except_table sections. We may support more
        types of exception handling information in the future.

        :param dwarf:   The DWARF info object from pyelftools.
        :return:        None
        """

        try:
            lsda = LSDAExceptionTable(self._binary_stream, self.arch.bits, self.arch.memory_endness == "Iend_LE")
            for entry in dwarf.EH_CFI_entries():
                if (
                    isinstance(entry, callframe.FDE)
                    and hasattr(entry, "lsda_pointer")
                    and entry.lsda_pointer is not None
                ):
                    # function address
                    func_addr = entry.header.get("initial_location", None)
                    if func_addr is None:
                        log.warning(
                            "Unexpected FDE structure: "
                            "initial_location is not specified while LSDA pointer is available."
                        )
                        continue

                    # Load and parse LSDA exception table
                    file_offset = self.addr_to_offset(entry.lsda_pointer)
                    lsda_exception_table = lsda.parse_lsda(entry.lsda_pointer, file_offset)
                    for exc in lsda_exception_table:
                        handling = ExceptionHandling(
                            func_addr + exc.cs_start,
                            exc.cs_len,
                            handler_addr=func_addr + exc.cs_lp if exc.cs_lp != 0 else None,
                            func_addr=func_addr,
                        )
                        self.exception_handlings.append(handling)

        except (DWARFError, ValueError):
            log.warning("An exception occurred in pyelftools when loading FDE information.", exc_info=True)

    def _load_line_info(self, dwarf):
        """
        Generates addr_to_line as a mapping: addr -> (filename, lineno).
        Lineno is one-indexed.
        """
        for cu in dwarf.iter_CUs():
            comp_dir = "."
            try:
                die = cu.get_top_DIE()
            except KeyError:
                # pyelftools is not very resilient
                continue
            if "DW_AT_comp_dir" in die.attributes:
                comp_dir = die.attributes["DW_AT_comp_dir"].value.decode()
            try:
                lineprog = dwarf.line_program_for_CU(cu)
            except ELFParseError:
                continue
            if lineprog is None:
                continue
            file_cache = {}
            for line in lineprog.get_entries():
                if line.state is None:
                    continue
                if line.state.file in file_cache:
                    filename = file_cache[line.state.file]
                else:
                    file_entry = lineprog.header["file_entry"][line.state.file - 1]
                    if file_entry["dir_index"] == 0:
                        filename = pathlib.PurePosixPath(comp_dir) / file_entry.name.decode()
                    else:
                        filename = (
                            pathlib.PurePosixPath(comp_dir)
                            / lineprog.header["include_directory"][file_entry["dir_index"] - 1].decode()
                            / file_entry.name.decode()
                        )
                    file_cache[line.state.file] = filename

                relocated_addr = AT.from_lva(line.state.address, self).to_mva()
                if relocated_addr not in self.addr_to_line:
                    self.addr_to_line[relocated_addr] = set()
                self.addr_to_line[relocated_addr].add((str(filename), line.state.line))

    @staticmethod
    def _load_low_high_pc_form_die(die: DIE):
        """
        Load low and high pc from a DIE.

        :param die:     The DIE object from pyelftools.
        :return:        low_pc, high_pc
        """
        if "DW_AT_low_pc" not in die.attributes:
            return None, None
        lowpc = die.attributes["DW_AT_low_pc"].value

        if "DW_AT_high_pc" not in die.attributes:
            return lowpc, None

        # DWARF v4 in section 2.17 describes how to interpret the
        # DW_AT_high_pc attribute based on the class of its form.
        # For class 'address' it's taken as an absolute address
        # (similarly to DW_AT_low_pc); for class 'constant', it's
        # an offset from DW_AT_low_pc.
        highpc_attr = die.attributes["DW_AT_high_pc"]
        highpc_attr_class = describe_form_class(highpc_attr.form)
        if highpc_attr_class == "address":
            highpc = highpc_attr.value
        elif highpc_attr_class == "constant":
            highpc = lowpc + highpc_attr.value
        else:
            log.warning("Error: invalid DW_AT_high_pc class:%s", highpc_attr_class)
            return lowpc, None
        return lowpc, highpc

    def _load_dies(self, dwarf: DWARFInfo):
        """
        Load DIEs and CUs from DWARF.

        :param dwarf:   The DWARF info object from pyelftools.
        :return:        None
        """
        compilation_units: list[CompilationUnit] = []
        type_list: dict[int, VariableType] = {}

        for cu in dwarf.iter_CUs():
            expr_parser = DWARFExprParser(cu.structs)

            # scan the whole die tree for DW_TAG_base_type
            try:
                for die in cu.iter_DIEs():
                    if VariableType.supported_die(die):
                        var_type = VariableType.read_from_die(die, self)
                        if var_type is not None:
                            type_list[die.offset] = var_type
            except KeyError:
                # pyelftools is not very resilient - we need to catch KeyErrors here
                continue

            top_die = cu.get_top_DIE()

            if top_die.tag != "DW_TAG_compile_unit":
                log.warning("ignore a top die with unexpected tag")
                continue

            die_name = top_die.attributes.get("DW_AT_name", None)
            die_comp_dir = top_die.attributes.get("DW_AT_comp_dir", None)
            die_low_pc, die_high_pc = self._load_low_high_pc_form_die(top_die)
            die_lang = top_die.attributes.get("DW_AT_language", None)

            if (
                die_name is None
                or die_comp_dir is None
                or die_low_pc is None
                or die_high_pc is None
                or die_lang is None
            ):
                continue

            die_name = die_name.value.decode("utf-8")
            die_comp_dir = die_comp_dir.value.decode("utf-8")
            die_lang = describe_attr_value(die_lang, top_die, top_die.offset)

            cu_ = CompilationUnit(die_name, die_comp_dir, die_low_pc, die_high_pc, die_lang, self)
            compilation_units.append(cu_)

            for die_child in cu.iter_DIE_children(top_die):
                if die_child.tag == "DW_TAG_variable":
                    # load global variable
                    var = Variable.from_die(die_child, expr_parser, self)
                    var.decl_file = cu_.file_path
                    cu_.global_variables.append(var)
                elif die_child.tag == "DW_TAG_subprogram":
                    # load subprogram
                    sub_prog = self._load_die_lex_block(die_child, expr_parser, type_list, cu, cu_.file_path, None)
                    if sub_prog is not None:
                        cu_.functions[sub_prog.low_pc] = sub_prog

        self.type_list = type_list
        self.compilation_units = compilation_units

    def _load_die_lex_block(self, die: DIE, expr_parser, type_list, cu, file_path, subprogram) -> LexicalBlock:
        if "DW_AT_name" in die.attributes:
            name = die.attributes["DW_AT_name"].value.decode("utf-8")
        else:
            name = None

        low_pc, high_pc = self._load_low_high_pc_form_die(die)
        if low_pc is None or high_pc is None:
            return None

        if subprogram is None:
            subprogram = block = Subprogram(name, low_pc, high_pc)
        else:
            block = LexicalBlock(low_pc, high_pc)

        for sub_die in cu.iter_DIE_children(die):
            if sub_die.tag in ["DW_TAG_variable", "DW_TAG_formal_parameter"]:
                # load local variable
                var = Variable.from_die(sub_die, expr_parser, self, block)
                var.decl_file = file_path
                subprogram.local_variables.append(var)
            elif sub_die.tag == "DW_TAG_lexical_block":
                sub_block = self._load_die_lex_block(sub_die, expr_parser, type_list, cu, file_path, subprogram)
                if sub_block is not None:
                    block.child_blocks.append(sub_block)

        return block

    #
    # Private Methods... really. Calling these out of context
    # will probably break things. Caveat emptor.
    #

    def __register_segments(self):
        self.linking = "static"
        if self.is_relocatable and self._reader.num_segments() > 0:
            # WTF?
            raise CLEError("Relocatable objects with segments are not supported.")

        type_to_seg_mapping = defaultdict(list)
        for seg_readelf in self._reader.iter_segments():
            type_to_seg_mapping[seg_readelf.header.p_type].append(seg_readelf)

        # PT_LOAD segments must be processed first so that the memory_backers for the other segments exist
        for seg in type_to_seg_mapping["PT_LOAD"]:
            self._load_segment(seg)

        # the order of processing for the other three handled segment_types should not matter, but let's have
        # it consistent
        for seg in type_to_seg_mapping["PT_DYNAMIC"]:
            self.__register_dyn(seg)
            self.linking = "dynamic"

        for seg in type_to_seg_mapping["PT_TLS"]:
            self.__register_tls(seg)

        for seg in type_to_seg_mapping["PT_GNU_STACK"]:
            self.execstack = bool(seg.header.p_flags & 1)

        for seg in type_to_seg_mapping["PT_GNU_RELRO"]:
            self.__register_relro(seg)

    def __register_dyn(self, seg_readelf):
        """
        Parse the dynamic section for dynamically linked objects.
        """
        # PATHOLOGICAL CASE
        # some elf files have a dyn with filesz = 0 but actually do contain content. this is valid. apparently.
        # this is a hack. there is certainly a better way to do this
        # ref (for fish and audrey's eyes only) https://shellphish.slack.com/archives/D08QXM76Y/p1649197435856279
        if seg_readelf.header.p_filesz == 0 and seg_readelf.header.p_memsz != 0:
            seg_readelf.header.p_filesz = seg_readelf.header.p_memsz
            seg_readelf.header.p_offset = AT.from_lva(seg_readelf.header.p_vaddr, self).to_rva()
            dynamic.Dynamic.__init__(
                seg_readelf,
                self.memory,  # YIKES
                seg_readelf.elffile,
                seg_readelf._stringtable,
                seg_readelf.header.p_offset,
                False,
            )

        runpath, rpath = "", ""
        for tag in seg_readelf.iter_tags():
            # Create a dictionary, self._dynamic, mapping DT_* strings to their values
            tagstr = self.arch.translate_dynamic_tag(tag.entry.d_tag)
            self._dynamic[tagstr] = tag.entry.d_val
            # For tags that may appear more than once, handle them here
            if tagstr == "DT_NEEDED":
                self.deps.append(maybedecode(tag.needed))
            elif tagstr == "DT_SONAME":
                self.provides = maybedecode(tag.soname)
            elif tagstr == "DT_RUNPATH":
                runpath = maybedecode(tag.runpath)
            elif tagstr == "DT_RPATH":
                rpath = maybedecode(tag.rpath)

        self.extra_load_path = self.__parse_rpath(runpath, rpath)

        strtab = seg_readelf._get_stringtable()
        if strtab is None:
            log.warning("Unexpected return value from pyelftools: stringtable object is None.")
            return
        self.__neuter_streams(strtab)

        # To extract symbols from binaries without section headers, we need to hack into pyelftools.
        # TODO: pyelftools is less bad than it used to be. how much of this can go away?
        # None of the following things make sense without a string table or a symbol table
        if "DT_STRTAB" in self._dynamic and "DT_SYMTAB" in self._dynamic and "DT_SYMENT" in self._dynamic:
            num_symbols = seg_readelf.num_symbols()  # this is not actually reliable

            # Construct our own symbol table to hack around pyreadelf assuming section headers are around
            entsize = self._dynamic["DT_SYMENT"]
            fakesymtabheader = {
                "sh_offset": AT.from_lva(self._dynamic["DT_SYMTAB"], self).to_rva(),
                "sh_entsize": entsize,
                "sh_size": entsize * num_symbols,
                "sh_flags": 0,
                "sh_addralign": 0,
            }
            dynsym = elffile.SymbolTableSection(fakesymtabheader, "symtab_cle", self._reader, strtab)
            dynsym.stream = self.memory
            dynsym.elffile = None

            # set up the hash table, preferring the gnu hash section to the old hash section
            # the hash table lets you get any symbol given its name
            if "DT_GNU_HASH" in self._dynamic:
                self.hashtable = GNUHashTable(
                    dynsym, self.memory, AT.from_lva(self._dynamic["DT_GNU_HASH"], self).to_rva(), self.arch
                )
            elif "DT_HASH" in self._dynamic:
                self.hashtable = ELFHashTable(
                    dynsym, self.memory, AT.from_lva(self._dynamic["DT_HASH"], self).to_rva(), self.arch
                )
            else:
                log.warning("No hash table available in %s", self.binary)

            # SYMBOL VERSIONING
            # REJOICE
            if "DT_VERNEED" in self._dynamic and "DT_VERNEEDNUM" in self._dynamic:
                verneed_count = self._dynamic["DT_VERNEEDNUM"]
                verneed_entsize = self._reader.structs.Elf_Verneed.sizeof()
                fake_verneed_header = {
                    "sh_offset": AT.from_lva(self._dynamic["DT_VERNEED"], self).to_rva(),
                    "sh_entsize": verneed_entsize,
                    "sh_size": verneed_entsize * verneed_count,
                    "sh_flags": 0,
                    "sh_addralign": 0,
                    "sh_info": verneed_count,
                }
                readelf_verneed = elffile.GNUVerNeedSection(fake_verneed_header, "verneed_cle", self._reader, strtab)
                for _, aux in readelf_verneed.iter_versions():
                    for vaux in aux:
                        self._versions[vaux.entry.vna_other] = vaux.name

            if "DT_VERDEF" in self._dynamic and "DT_VERDEFNUM" in self._dynamic:
                verdef_count = self._dynamic["DT_VERDEFNUM"]
                verdef_entsize = self._reader.structs.Elf_Verdef.sizeof()
                fake_verdef_header = {
                    "sh_offset": AT.from_lva(self._dynamic["DT_VERDEF"], self).to_rva(),
                    "sh_entsize": verdef_entsize,
                    "sh_size": verdef_entsize * verdef_count,
                    "sh_flags": 0,
                    "sh_addralign": 0,
                    "sh_info": verdef_count,
                }
                readelf_verdef = elffile.GNUVerDefSection(fake_verdef_header, "verdef_cle", self._reader, strtab)
                for ver, aux in readelf_verdef.iter_versions():
                    for vaux in aux:
                        self._versions[ver.entry.vd_ndx] = vaux.name
                        break  # not a typo - only the first aux entry has anything important (?)

            if "DT_VERSYM" in self._dynamic:
                versym_entsize = self._reader.structs.Elf_Versym.sizeof()
                fake_versym_header = {
                    "sh_offset": AT.from_lva(self._dynamic["DT_VERSYM"], self).to_rva(),
                    "sh_entsize": versym_entsize,
                    "sh_size": versym_entsize * num_symbols,
                    "sh_flags": 0,
                    "sh_addralign": 0,
                }
                readelf_versym = elffile.GNUVerSymSection(fake_versym_header, "versym_cle", self._reader, dynsym)
                readelf_versym.stream = self.memory
                readelf_versym.elffile = None
                self._vertable = readelf_versym

            # mips' relocations are absolutely screwed up, handle some of them here.
            self.__relocate_mips(dynsym)

            # perform a lot of checks to figure out what kind of relocation tables are around
            self.rela_type = None
            if "DT_PLTREL" in self._dynamic:
                if self._dynamic["DT_PLTREL"] == 7:
                    self.rela_type = "RELA"
                    relentsz = self._reader.structs.Elf_Rela.sizeof()
                elif self._dynamic["DT_PLTREL"] == 17:
                    self.rela_type = "REL"
                    relentsz = self._reader.structs.Elf_Rel.sizeof()
                else:
                    raise CLEInvalidBinaryError("DT_PLTREL is not REL or RELA?")
            else:
                if "DT_RELA" in self._dynamic:
                    self.rela_type = "RELA"
                    relentsz = self._reader.structs.Elf_Rela.sizeof()
                elif "DT_REL" in self._dynamic:
                    self.rela_type = "REL"
                    relentsz = self._reader.structs.Elf_Rel.sizeof()
                else:
                    return

            # try to parse relocations out of a table of type DT_REL{,A}
            rela_tag = "DT_" + self.rela_type
            relsz_tag = rela_tag + "SZ"
            if rela_tag in self._dynamic:
                reloffset = AT.from_lva(self._dynamic[rela_tag], self).to_rva()
                if relsz_tag not in self._dynamic:
                    raise CLEInvalidBinaryError(f"Dynamic section contains {rela_tag} but not {relsz_tag}")
                relsz = self._dynamic[relsz_tag]
                fakerelheader = {
                    "sh_offset": reloffset,
                    "sh_type": "SHT_" + self.rela_type,
                    "sh_entsize": relentsz,
                    "sh_size": relsz,
                    "sh_flags": 0,
                    "sh_addralign": 0,
                }
                readelf_relocsec = elffile.RelocationSection(fakerelheader, "reloc_cle", self._reader)
                # support multiple versions of pyelftools
                readelf_relocsec.stream = self.memory
                readelf_relocsec._stream = self.memory
                readelf_relocsec.elffile = None
                self.__register_relocs(readelf_relocsec, dynsym)

            # try to parse relocations out of a table of type DT_JMPREL
            if "DT_JMPREL" in self._dynamic:
                jmpreloffset = AT.from_lva(self._dynamic["DT_JMPREL"], self).to_rva()
                if "DT_PLTRELSZ" not in self._dynamic:
                    raise CLEInvalidBinaryError("Dynamic section contains DT_JMPREL but not DT_PLTRELSZ")
                jmprelsz = self._dynamic["DT_PLTRELSZ"]
                fakejmprelheader = {
                    "sh_offset": jmpreloffset,
                    "sh_type": "SHT_" + self.rela_type,
                    "sh_entsize": relentsz,
                    "sh_size": jmprelsz,
                    "sh_flags": 0,
                    "sh_addralign": 0,
                }
                readelf_jmprelsec = elffile.RelocationSection(fakejmprelheader, "jmprel_cle", self._reader)
                # support multiple versions of pyelftools
                readelf_jmprelsec.stream = self.memory
                readelf_jmprelsec._stream = self.memory
                readelf_jmprelsec.elffile = None
                self.__register_relocs(readelf_jmprelsec, dynsym, force_jmprel=True)

            self.__register_section_symbols(dynsym)

    def __parse_rpath(self, runpath, rpath):
        """
        Split RPATH/RUNPATH tags into a list of paths while expanding rpath tokens.
        """
        # DT_RUNPATH takes predence over DT_RPATH
        if len(runpath) > 0:
            pass
        else:
            runpath = rpath

        parts = []
        for part in runpath.split(":"):
            # does not handle $LIB/$PLATFORM substitutions yet
            if self.binary is not None:
                part = part.replace("$ORIGIN", os.path.dirname(self.binary))
            elif "$ORIGIN" in part:
                log.warning(
                    "DT_RUNPATH/DT_RPATH of the binary contains $ORIGIN tokens but no self.binary, "
                    "some libraries might be not found"
                )
            parts.append(part)
        return parts

    def __register_relocs(self, section, dynsym=None, force_jmprel=False):
        got_min = got_max = 0

        if not force_jmprel:
            got_sec = self._reader.get_section_by_name(".got")
            if got_sec is not None:
                got_min = got_sec.header.sh_addr
                got_max = got_min + got_sec.header.sh_size

        if section.header["sh_offset"] in self.__parsed_reloc_tables:
            return
        self.__parsed_reloc_tables.add(section.header["sh_offset"])

        # Get the target section's remapping offset for relocations
        dest_sec = None
        dest_sec_idx = section.header.get("sh_info", None)
        if dest_sec_idx is not None:
            try:
                dest_sec = self.sections[dest_sec_idx]
            except IndexError:
                log.warning("the relocation section %s refers to unknown section index: %d", section.name, dest_sec_idx)
            else:
                if dest_sec.is_active and not dest_sec.occupies_memory:
                    # The target section is not loaded into memory, so just continue
                    return

        symtab = self._reader.get_section(section.header["sh_link"]) if "sh_link" in section.header else dynsym
        if isinstance(symtab, elftools.elf.sections.NullSection):
            # Oh my god Atmel please stop
            symtab = self._reader.get_section_by_name(".symtab")
        relocs = []
        for readelf_reloc in section.iter_relocations():
            # MIPS64 is just plain old fucked up
            # https://www.sourceware.org/ml/libc-alpha/2003-03/msg00153.html
            if self.arch.name == "MIPS64":
                if not hasattr(readelf_reloc.entry, "r_info_type2") and hasattr(readelf_reloc.entry, "r_info_type3"):
                    raise CLECompatibilityError(
                        "This code relies on `pyelftools` features that are not available on versions 0.26 and below."
                    )

                type_1 = readelf_reloc.entry.r_info_type
                type_2 = readelf_reloc.entry.r_info_type2
                type_3 = readelf_reloc.entry.r_info_type3
                symbol = self.get_symbol(readelf_reloc.entry.r_info_sym, symtab)

                if type_1 != 0:
                    readelf_reloc.entry.r_info_type = type_1
                    reloc = self._make_reloc(readelf_reloc, symbol, dest_sec)
                    if reloc is not None:
                        relocs.append(reloc)
                        self.relocs.append(reloc)
                if type_2 != 0:
                    readelf_reloc.entry.r_info_type = type_2
                    reloc = self._make_reloc(readelf_reloc, symbol, dest_sec)
                    if reloc is not None:
                        relocs.append(reloc)
                        self.relocs.append(reloc)
                if type_3 != 0:
                    readelf_reloc.entry.r_info_type = type_3
                    reloc = self._make_reloc(readelf_reloc, symbol, dest_sec)
                    if reloc is not None:
                        relocs.append(reloc)
                        self.relocs.append(reloc)
            else:
                symbol = self.get_symbol(readelf_reloc.entry.r_info_sym, symtab)
                if symbol is None:
                    continue
                reloc = self._make_reloc(readelf_reloc, symbol, dest_sec)
                if reloc is not None:
                    relocs.append(reloc)
                    self.relocs.append(reloc)

        for reloc in relocs:
            if reloc.symbol.name != "" and (force_jmprel or got_min <= reloc.linked_addr < got_max):
                self.jmprel[reloc.symbol.name] = reloc

    def __register_tls(self, seg_readelf):
        self.tls_block_size = seg_readelf.header.p_memsz
        self.tls_data_size = seg_readelf.header.p_filesz
        self.tls_data_start = AT.from_lva(seg_readelf.header.p_vaddr, self).to_rva()
        if self.tls_block_size != 0 or self.tls_data_size != 0:
            self.tls_used = True

    def __register_relro(self, segment_relro):
        segment_relro = ELFSegment(segment_relro, relro=True)
        vaddr = ALIGN_DOWN(segment_relro.vaddr, self.loader.page_size)
        vaddr_end = ALIGN_UP(vaddr + segment_relro.memsize, self.loader.page_size)
        vaddr_endfile = ALIGN_UP(vaddr + segment_relro.filesize, self.loader.page_size)

        segment_relro.offset = ALIGN_DOWN(segment_relro.offset, self.loader.page_size)
        segment_relro.vaddr = vaddr
        segment_relro.memsize = vaddr_end - vaddr
        segment_relro.filesize = vaddr_endfile - vaddr

        def ___segments_overlap(seg1, seg2):
            # Re-arrange so seg1 starts first
            seg1, seg2 = (seg1, seg2) if seg1.min_addr < seg2.min_addr else (seg2, seg1)
            # seg1 and seg2 overlap if seg2 starts before seg1 ends
            return seg2.min_addr <= seg1.max_addr

        overlapping_segments = [seg for seg in self.segments if ___segments_overlap(segment_relro, seg)]

        if len(overlapping_segments) == 0:
            log.error("RELRO segment does not overlap with any loaded segment.")
            return

        if len(overlapping_segments) > 1:
            # I don't think this ever happens.  If it does, weshould
            # probably also split the RELRO segment so that each one
            # has the right permissions in case the two overlapping
            # segments have different permissions.
            log.warning("RELRO segment overlaps multiple segments.")

        for overlapping_segment in overlapping_segments:
            # We will split the overlapping segment into two pieces:
            # one for the segment below the RELRO segment, and one for
            # above.
            segment_below = copy.copy(overlapping_segment)
            segment_below.memsize = segment_relro.min_addr - overlapping_segment.min_addr
            segment_below.filesize = max(segment_below.filesize, segment_below.memsize)

            segment_above = copy.copy(overlapping_segment)
            segment_above.vaddr = segment_relro.max_addr + 1
            segment_above.memsize = overlapping_segment.max_addr - segment_above.vaddr + 1
            segment_above.filesize = max(0, segment_above.filesize - segment_relro.memsize - segment_below.memsize)
            segment_above.offset = segment_above.offset + segment_below.memsize + segment_relro.memsize

            split_segments = [segment_below, segment_relro, segment_above]
            split_segments = [seg for seg in split_segments if seg.memsize > 0]

            # Remove the original segment
            self.segments.remove(overlapping_segment)

            # Add the new ones
            for seg in split_segments:
                self.segments.append(seg)

        # Add the actual relro segment, and mark it as always
        # read-only.  We use the flags of the overlapping original
        # segment.
        segment_relro.flags = overlapping_segments[0].flags & ~2

    def __register_sections(self):
        new_addr = 0
        sec_list = []

        for sec_readelf in self._reader.iter_sections():
            remap_offset = 0
            if self.is_relocatable and sec_readelf.header["sh_flags"] & 2:  # alloc flag
                # Relocatable objects' section addresses are meaningless (they are meant to be relocated anyway)
                # We thus have to map them manually to valid virtual addresses to emulate a linker's behaviour.
                sh_addr = sec_readelf.header["sh_addr"]
                align = sec_readelf.header["sh_addralign"]
                if align > 0:
                    new_addr = (new_addr + (align - 1)) // align * align

                remap_offset = new_addr - sh_addr
                new_addr += sec_readelf.header["sh_size"]  # address for next section

            section = ELFSection(sec_readelf, remap_offset=remap_offset)
            sec_list.append((sec_readelf, section))

            # Register sections first, process later - this is required by relocatable objects
            self.sections.append(section)
            self.sections_map[section.name] = section

        for sec_readelf, section in sec_list:
            if isinstance(sec_readelf, elffile.SymbolTableSection):
                self.__register_section_symbols(sec_readelf)
            if isinstance(sec_readelf, elffile.RelocationSection) and not (
                "DT_REL" in self._dynamic or "DT_RELA" in self._dynamic or "DT_JMPREL" in self._dynamic
            ):
                self.__register_relocs(sec_readelf, dynsym=None)

            if section.occupies_memory:  # alloc flag - stick in memory maybe!
                if (
                    AT.from_lva(section.vaddr, self).to_rva() not in self.memory
                ):  # only allocate if not already allocated (i.e. by program header)
                    if section.type == "SHT_NOBITS":
                        self.memory.add_backer(
                            AT.from_lva(section.vaddr, self).to_rva(),
                            b"\0" * sec_readelf.header["sh_size"],
                            overwrite=True,
                        )
                    elif section.type == "SHT_NOTE":
                        pass  # observed this case in angr/angr#3829
                    else:  # elif section.type == 'SHT_PROGBITS':
                        self.memory.add_backer(
                            AT.from_lva(section.vaddr, self).to_rva(), sec_readelf.data(), overwrite=True
                        )

            if sec_readelf.header.sh_type == "SHT_NOTE":
                self.__register_notes(sec_readelf)

            if section.name == ".comment":
                self.__analyze_comments(sec_readelf.data())

    def __register_notes(self, sec_readelf):
        for note in sec_readelf.iter_notes():
            if note.n_type == "NT_GNU_BUILD_ID" and note.n_name == "GNU":
                if self.build_id is not None and self.build_id != note.n_desc:
                    log.error("Mismatched build IDs present")
                self.build_id = note.n_desc

    def __analyze_comments(self, data):
        try:
            data = data.decode().split("\0")
        except UnicodeDecodeError:
            return

        for line in data:
            versions = [word for word in line.replace("(", " ").replace(")", " ").split() if "." in word]
            if not versions:
                continue
            versions.sort(key=len)
            version = versions[-1]
            lline = line.lower()
            if "clang" in lline:
                compiler = "clang"
            elif "gcc" in lline:
                compiler = "gcc"
            else:
                continue

            self.compiler = compiler, version

    def __register_section_symbols(self, sec_re):
        for sym_re in sec_re.iter_symbols():
            self.symbols.add(self.get_symbol(sym_re))

    def __relocate_mips(self, symtab):
        if "DT_MIPS_BASE_ADDRESS" not in self._dynamic:
            return False
        # The MIPS GOT is an array of addresses, simple as that.
        got_local_num = self._dynamic["DT_MIPS_LOCAL_GOTNO"]  # number of local GOT entries
        # a.k.a the index of the first global GOT entry
        symtab_got_idx = self._dynamic["DT_MIPS_GOTSYM"]  # index of first symbol w/ GOT entry
        symbol_count = self._dynamic["DT_MIPS_SYMTABNO"]
        gotaddr = AT.from_lva(self._dynamic["DT_PLTGOT"], self).to_rva()
        wordsize = self.arch.bytes
        for i in range(2, got_local_num):
            reloc = MipsLocalReloc(self, None, gotaddr + i * wordsize)
            self.relocs.append(reloc)

        for i in range(symbol_count - symtab_got_idx):
            symbol = self.get_symbol(i + symtab_got_idx, symtab)
            reloc = MipsGlobalReloc(self, symbol, gotaddr + (i + got_local_num) * wordsize)
            self.relocs.append(reloc)
            self.jmprel[symbol.name] = reloc
        return True

    def __neuter_streams(self, obj):
        if isinstance(obj, elftools.elf.dynamic._DynamicStringTable):
            obj._stream = self.memory
            obj._table_offset = self._offset_to_rva(obj._table_offset)
        elif isinstance(obj, elftools.elf.sections.Section):
            if obj.header.sh_type == "SHT_NOBITS":
                obj.stream = None
                obj.elffile = None
                obj.header.sh_offset = None
            else:
                obj.stream = self.memory
                obj.elffile = None
                obj.header.sh_offset = self._offset_to_rva(obj.header.sh_offset)
        else:
            raise TypeError(f"Can't convert {type(obj)!r}")

    def _offset_to_rva(self, offset):
        return AT.from_mva(self.offset_to_addr(offset), self).to_rva()

    def __process_debug_file(self, filename):
        with open(filename, "rb") as fp:
            try:
                elf = elffile.ELFFile(fp)
            except ELFError:
                log.warning("pyelftools failed to load debug file %s", filename, exc_info=True)
                return

            for sec_readelf in elf.iter_sections():
                if isinstance(sec_readelf, elffile.SymbolTableSection):
                    self.__register_section_symbols(sec_readelf)
                elif sec_readelf.header.sh_type == "SHT_NOTE":
                    self.__register_notes(sec_readelf)

            has_dwarf_info = bool(elf.has_dwarf_info())
            if has_dwarf_info:
                try:
                    dwarf = elf.get_dwarf_info()
                except ELFError:
                    log.warning(
                        "An exception occurred in pyelftools when loading the DWARF information on %s.",
                        filename,
                        exc_info=True,
                    )
                    dwarf = None

                # debug symbols don't have eh_frame ever from what I can tell
                if dwarf:
                    self._load_line_info(dwarf)

    @staticmethod
    def _get_pcode_elf_opinions():
        """
        Load each .opinion file and gather all ELF constraints
        """
        if pypcode is None:
            raise CLEError("pypcode is not installed")

        def flatten_constraints(node):
            constraints = []
            for child in node.findall("constraint"):
                constraints.extend(flatten_constraints(child))
            attribs = node.attrib.copy()
            if not constraints:  # Leaf
                return [attribs]
            for c in constraints:  # Apply attribs to all children
                c.update(attribs)
            return constraints

        log.info("Loading opinions...")
        SPECFILES_DIR = pypcode.SPECFILES_DIR
        elf_opinions = []
        for archname in os.listdir(SPECFILES_DIR):
            langdir = os.path.join(SPECFILES_DIR, archname, "data", "languages")
            if not (os.path.exists(langdir) and os.path.isdir(langdir)):
                continue
            for filename in os.listdir(langdir):
                if not filename.endswith(".opinion"):
                    continue
                path = os.path.join(langdir, filename)
                opinion = xml.etree.ElementTree.parse(path).getroot()
                for c in opinion.findall("constraint"):
                    if c.attrib["loader"] == "Executable and Linking Format (ELF)":
                        elf_opinions.extend(flatten_constraints(c))
        return elf_opinions

    @staticmethod
    def _get_compatible_pcode_languages(reader):
        """
        Find compatible pypcode languages for this ELF file.
        """
        if pypcode is None:
            raise CLEError("pypcode is not installed")

        e_machine = reader.header.e_machine
        if isinstance(e_machine, str):
            e_machine = enums.ENUM_E_MACHINE[e_machine]

        endian = {"ELFDATANONE": None, "ELFDATA2LSB": "little", "ELFDATA2MSB": "big"}[reader.header.e_ident.EI_DATA]

        opinions = []
        for o in ELF._get_pcode_elf_opinions():
            if "endian" in o and o["endian"] != endian:
                continue
            if e_machine not in {int(p) for p in o["primary"].split(",")}:
                continue
            if "secondary" in o:
                try:
                    value = int(o["secondary"])
                    if value != reader.header.e_type:
                        continue
                except ValueError:
                    # FIXME: Mask parsing (spaces and DC 0b .... ..1. ..0.)
                    pass
            opinions.append(o)

        log.info("Available opinions: %s", opinions)

        languages = []
        for arch in pypcode.Arch.enumerate():
            for lang in arch.languages:
                for o in opinions:
                    if (reader.elfclass == 32 and int(lang.size) > 32) or (
                        reader.elfclass == 64 and int(lang.size) <= 32
                    ):
                        continue
                    if all(k not in lang.ldef.attrib or lang.ldef.attrib[k] == v for k, v in o.items()):
                        languages.append(lang)

        log.info("Found candidate languages: %s", [lang.id for lang in languages])
        return languages


register_backend("elf", ELF)



"""Relocation types for the S390X architecture.

Reference: https://github.com/IBM/s390x-abi/releases/download/v1.6.1/lzsabi_s390x.pdf pages 51-52
"""

from __future__ import annotations

from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericRelativeReloc,
    GenericTLSOffsetReloc,
)


class R_390_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_390_JMP_SLOT(GenericJumpslotReloc):
    pass


class R_390_RELATIVE(GenericRelativeReloc):
    pass


class R_390_64(GenericAbsoluteAddendReloc):
    pass


class R_390_TLS_TPOFF(GenericTLSOffsetReloc):
    pass


class R_390_IRELATIVE(GenericIRelativeReloc):
    pass


class R_390_COPY(GenericCopyReloc):
    pass


relocation_table_s390x = {
    # 1: R_390_8,
    # 2: R_390_12,
    # 3: R_390_16,
    # 4: R_390_32,
    # 5: R_390_PC32,
    # 6: R_390_GOT12,
    # 7: R_390_GOT32,
    # 8: R_390_PLT32,
    9: R_390_COPY,
    10: R_390_GLOB_DAT,
    11: R_390_JMP_SLOT,
    12: R_390_RELATIVE,
    # 13: R_390_GOTOFF32,
    # 14: R_390_GOTPC,
    # 15: R_390_GOT16,
    # 16: R_390_PC16,
    # 17: R_390_PC16DBL,
    # 18: R_390_PLT16DBL,
    # 19: R_390_PC32DBL,
    # 20: R_390_PLT32DBL,
    # 21: R_390_GOTPCDBL,
    22: R_390_64,
    # 23: R_390_PC64,
    # 24: R_390_GOT64,
    # 25: R_390_PLT64,
    # 26: R_390_GOTENT,
    # 27: R_390_GOTOFF16,
    # 28: R_390_GOTOFF64,
    # 29: R_390_GOTPLT12,
    # 30: R_390_GOTPLT16,
    # 31: R_390_GOTPLT32,
    # 32: R_390_GOTPLT64,
    # 33: R_390_GOTPLTENT,
    # 34: R_390_PLTOFF16,
    # 35: R_390_PLTOFF32,
    # 36: R_390_PLTOFF64,
    # 37: R_390_TLS_LOAD,
    # 38: R_390_TLS_GDCALL,
    # 39: R_390_TLS_LDCALL,
    # No 40 in doc
    # 41: R_390_TLS_GD64,
    # 42: R_390_TLS_GOTIE12,
    # No 43 in doc
    # 44: R_390_TLS_GOTIE64,
    # No 45 in doc
    # 46: R_390_TLS_LDM64,
    # No 47 in doc
    # 48: R_390_TLS_IE64,
    # 49: R_390_TLS_IEENT,
    # No 50 in doc
    # 51: R_390_TLS_LE64,
    # No 52 in doc
    # 53: R_390_TLS_LDO64,
    # 54: R_390_TLS_DTPMOD,
    # 55: R_390_TLS_DTPOFF,
    56: R_390_TLS_TPOFF,
    # 57: R_390_20,
    # 58: R_390_GOT20,
    # 59: R_390_GOTPLT20,
    # 60: R_390_TLS_GOTIE20,
    61: R_390_IRELATIVE,
    # 62: R_390_PC12DBL,
    # 63: R_390_PLT12DBL,
    # 64: R_390_PC24DBL,
    # 65: R_390_PLT24DBL,
}

__all__ = ("relocation_table_s390x",)



# This file is part of Mach-O Loader for CLE.
# Contributed December 2016 by Fraunhofer SIT (https://www.sit.fraunhofer.de/en/).
from __future__ import annotations

from cle.backends.region import Section

from .segment import MachOSegment

TYPE_MASK = 0x000000FF
ATTRIBUTES_MASK = 0xFFFFFF00


class MachOSection(Section):
    """
    Mach-O Section, only defined within the context of a Mach-O Segment.

        - offset is the offset into the file the region starts
        - vaddr (or just addr) is the virtual address
        - filesize (or just size) is the size of the region in the file
        - memsize (or vsize) is the size of the region when loaded into memory
        - segname is the corresponding segment's name without padding
        - sectname is the section's name without padding
        - align is the sections alignment as a power of 2
        - reloff is the file offset to the section's relocation entries
        - nreloc is the number of relocation entries for this section
        - flags is a bit vector containing per-section flags
        - r1 and r2 are values for the reserved1 and reserved2 fields respectively
    """

    def __init__(
        self,
        offset,
        vaddr,
        size,
        vsize,
        segname,
        sectname,
        align,
        reloff,
        nreloc,
        flags,
        r1,
        r2,
        parent_segment: MachOSegment | None = None,
    ):
        super().__init__(sectname.decode(), offset, vaddr, size)
        self.filesize = size
        self.memsize = vsize
        self.segname = segname.decode()
        self.sectname = sectname.decode()
        self.align = align
        self.reloff = reloff
        self.nreloc = nreloc
        self.flags = flags
        self.reserved1 = r1
        self.reserved2 = r2
        self.parent_segment = parent_segment

    @property
    def type(self):
        return self.flags & TYPE_MASK

    @property
    def attributes(self):
        return self.flags & ATTRIBUTES_MASK

    @property
    def is_readable(self):
        """
        Always true, because sections should always be readable
        :return:
        """
        return True

    @property
    def is_writable(self):
        """
        Returns the permission of the parent segment, because MachO sections simply inherit that
        :return:
        """
        return self.parent_segment.is_writable

    @property
    def is_executable(self):
        """
        Returns the permission of the parent segment, because MachO sections simply inherit that
        :return:
        """
        return self.parent_segment.is_executable

    @property
    def only_contains_uninitialized_data(self):
        """
        I actually don't know if this is true, but it seems like a saner assumption than true
        :return:
        """
        return False

    def __repr__(self):
        return "<Section: {} (part of Segment: {})| offset {:#x}, vaddr {:#x}, size {:#x}>".format(
            self.sectname if self.sectname else "Unnamed", self.segname, self.offset, self.vaddr, self.memsize
        )



from __future__ import annotations

import binascii
import logging
import re
import struct

import archinfo

from cle.errors import CLEError

from .backend import Backend, register_backend
from .region import Section, Segment

log = logging.getLogger(name=__name__)

__all__ = ("Hex",)

intel_hex_re = re.compile(
    b":([0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F][0-9a-fA-F][0-9a-fA-F])"
    b"([0-9a-fA-F][0-9a-fA-F])([0-9a-fA-F][0-9a-fA-F]+)*([0-9a-fA-F][0-9a-fA-F])"
)

HEX_TYPE_DATA = 0x00
HEX_TYPE_EOF = 0x01
HEX_TYPE_EXTSEGADDR = 0x02
HEX_TYPE_STARTSEGADDR = 0x03
HEX_TYPE_EXTLINEARADDR = 0x04
HEX_TYPE_STARTLINEARADDR = 0x05


class HexSection(Section):
    """
    Describes a section in memory after loading a HEX file into memory. This section may come from the HEX file or
    added separately (e.g., the MMIO region).
    """

    def __init__(self, name, vaddr, size, initialized=True, readable=True, writable=False, executable=True):
        super().__init__(name, 0, vaddr, size)
        self.initialized = initialized
        self.readable = readable
        self.writable = writable
        self.executable = executable

    @property
    def is_readable(self):
        return self.readable

    @property
    def is_writable(self):
        return self.writable

    @property
    def is_executable(self):
        return self.executable

    @property
    def only_contains_uninitialized_data(self) -> bool:
        return not self.initialized


class Hex(Backend):
    """
    A loader for Intel Hex Objects
    See https://en.wikipedia.org/wiki/Intel_HEX
    """

    is_default = True  # Tell CLE to automatically consider using the Hex backend

    @staticmethod
    def parse_record(line):
        m = intel_hex_re.match(line)
        if not m:
            raise CLEError(f"Invalid HEX record: {line}")
        my_cksum = 0
        count, addr, rectype, data, cksum = m.groups()
        cksum = int(cksum, 16)
        for d in binascii.unhexlify(line[1:-2]):
            my_cksum = (my_cksum + d) % 256
        my_cksum = ((my_cksum ^ 0xFF) + 1) % 256
        if my_cksum != cksum:
            raise CLEError(f"Invalid checksum: Computed {hex(my_cksum)}, found {hex(cksum)}")
        count = int(count, 16)
        addr = int(addr, 16)
        rectype = int(rectype, 16)
        if data:
            data = binascii.unhexlify(data)
        if data and count != len(data):
            raise CLEError("Data length field does not match length of actual data: " + line)
        return rectype, addr, data

    @staticmethod
    def coalesce_regions(regions):
        # Lots of tiny memory regions is bad!
        # The greedy algorithm to smash them together:
        result = []
        last_addr: int | None = None
        last_data: list[bytes] | None = None
        last_size: int | None = None
        for addr, region in sorted(regions):
            if last_addr is not None and last_addr + last_size == addr:
                last_data.append(region)
                last_size += len(region)
            else:
                if last_addr is not None:
                    result.append((last_addr, b"".join(last_data)))
                last_addr, last_data, last_size = addr, [region], len(region)

        if last_addr is not None:
            result.append((last_addr, b"".join(last_data)))

        return result

    def __init__(
        self, *args, ignore_missing_arch: bool = False, extra_regions: list[dict[str, int | bool]] = None, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.set_load_args(ignore_missing_arch=ignore_missing_arch)
        if extra_regions:
            self.set_load_args(extra_regions=extra_regions)

        if self._arch is None:
            if ignore_missing_arch:
                # used internally for testing. we use a default architecture
                self.set_arch(archinfo.arch_from_id("amd64"))
            else:
                raise CLEError(
                    "To use the Hex binary backend, you need to specify an architecture in the loader options."
                )

        # Do the whole thing in one shot.
        self.os = "unknown"
        got_base = False
        got_entry = False
        self._binary_stream.seek(0)
        string = self._binary_stream.read()
        recs = string.splitlines()
        regions = []
        max_addr = 0
        min_addr = 0xFFFFFFFFFFFFFFFF
        self._base_address = 0
        for rec in recs:
            rectype, addr, data = Hex.parse_record(rec)
            if rectype == HEX_TYPE_DATA:
                addr += self._base_address
                # l.debug("Loading %d bytes at " % len(data) + hex(addr))
                # Raw data.  Put the bytes
                regions.append((addr, data))
                # We have to be careful about the min and max addrs
                if addr < min_addr:
                    min_addr = addr
                max_addr = max(max_addr, addr + len(data) - 1)
            elif rectype == HEX_TYPE_EOF:
                # EOF
                log.debug("Got EOF record.")
                break
            elif rectype == HEX_TYPE_EXTSEGADDR:
                # "Extended Mode" Segment address, take this value, multiply by 16, make the base
                self._base_address = struct.unpack(">H", data)[0] * 16
                got_base = True
                log.debug("Loading a segment at %#x", self._base_address)
            elif rectype == HEX_TYPE_STARTSEGADDR:
                # Four bytes, the segment and the initial IP
                got_base = True
                got_entry = True
                self._initial_cs, self._initial_ip = struct.unpack(">HH", data)
                # The whole thing is the entry, as far as angr is concerned.
                self._entry = struct.unpack(">I", data)[0]
                log.debug("Got entry point at %#x", self._entry)
            elif rectype == HEX_TYPE_EXTLINEARADDR:
                got_base = True
                # Specifies the base for all future data bytes.
                self._base_address = struct.unpack(">H", data)[0] << 16
                log.debug("Loading a segment at %#x", self._base_address)
            elif rectype == HEX_TYPE_STARTLINEARADDR:
                got_entry = True
                # The 32-bit EIP, really the same as STARTSEGADDR, but some compilers pick one over the other.
                self._entry = struct.unpack(">I", data)[0]
                log.debug("Found entry point at %#x", self._entry)
                self._initial_eip = self._entry
            else:
                raise CLEError("This HEX Object type is not implemented: " + hex(rectype))
        if not got_base:
            log.warning("No base address was found in this HEX object file. It is assumed to be 0")

        self.mapped_base = 0
        self._min_addr = min_addr
        self._max_addr = max_addr - self.mapped_base

        if self.load_args.get("entry_point") is not None:
            self._entry = self._custom_entry_point = self.load_args["entry_point"] - min_addr
            got_entry = True

        if not got_entry:
            log.warning(
                "No entry point was found in this HEX object file, and it is assumed to be 0. "
                "Specify one with `entry_point` to override."
            )

        # HEX specifies a ton of tiny little memory regions.  We now smash them together to make things faster.
        new_regions = Hex.coalesce_regions(regions)
        self.regions: list[tuple[int, int]] = []  # A list of (addr, size)
        for addr, data in new_regions:
            self.memory.add_backer(addr, data)
            self.regions.append((addr, len(data)))
            self.segments.append(Segment(0, addr, len(data), len(data)))
            self.sections.append(HexSection(f"sec_{addr:x}", addr, len(data)))

        self._create_extra_regions()

    @staticmethod
    def is_compatible(stream):
        stream.seek(0)
        s = stream.read(0x10)
        stream.seek(0)
        return s.startswith(b":")

    def _create_extra_regions(self) -> None:
        """
        Parse load_args["extra_regions"] and create segments/sections accordingly. This method should be called at the
        end of the loading process (if the backend supports loading extra regions).

        "extra_regions": [
        {
            "name": "ram",
            "addr": 0x20000000,
            "size": 0x1000,
            "readable": True,
            "writable": True,
            "executable": False,
        },
        {
            "name": "mmio",
            ...
        },
        ]
        """

        # in case special regions are specified
        if "extra_regions" not in self.load_args or not self.load_args["extra_regions"]:
            return
        for region in self.load_args["extra_regions"]:
            self.segments.append(Segment(0, region["addr"], 0, region["size"]))
            self.sections.append(
                HexSection(
                    region["name"],
                    region["addr"],
                    region["size"],
                    initialized=region.get("initialized", True),
                    readable=region.get("readable", True),
                    writable=region.get("writable", False),
                    executable=region.get("executable", True),
                )
            )


register_backend("hex", Hex)



from __future__ import annotations

import logging

from cle.backends.symbol import Symbol, SymbolType

log = logging.getLogger(name=__name__)


class WinSymbol(Symbol):
    """
    Represents a symbol for the PE format.
    """

    def __init__(
        self, owner, name, addr, is_import, is_export, ordinal_number, forwarder, symbol_type=SymbolType.TYPE_FUNCTION
    ):
        super().__init__(owner, name, addr, owner.arch.bytes, symbol_type)
        self.is_import = is_import
        self.is_export = is_export
        self.ordinal_number = ordinal_number
        self.forwarder = forwarder
        self.is_forward = forwarder is not None

    def resolve_forwarder(self):
        if self.resolvedby is not None:
            return self.resolvedby

        sym = self
        seen = set()
        while sym is not None and sym.is_forward and sym.forwarder is not None:  # FORWARDING
            if sym.forwarder in seen:
                log.warning("Infinite forwarding loop for %s", self)
                return None
            seen.add(sym.forwarder)
            owner_name, name = sym.forwarder.split(".", 1)
            owner = self.owner.loader.find_object(owner_name)
            if owner is None:
                return None
            sym = owner.get_symbol(name)

        self.resolvedby = sym
        return sym



from __future__ import annotations

import binascii
import logging
import re

import archinfo

from cle.errors import CLEError

from .backend import Backend, register_backend

log = logging.getLogger(name=__name__)

__all__ = ("SRec",)

srec_regex = "S([0-9])([0-9a-fA-F]{{2}})([0-9a-fA-F]{{{addr_size}}})([0-9a-fA-F]{{4,64}})([0-9a-fA-F]{{2}})"
SREC_ADDR_SIZE = {"0": 16, "1": 16, "5": 16, "9": 16, "2": 24, "6": 24, "8": 24, "3": 32, "7": 32}


SREC_HEADER = 0x00
SREC_DATA = {0x01, 0x02, 0x03}
SREC_START_EXEC = {0x7, 0x8, 0x9}
SREC_COUNT = {0x5, 0x6}
HEX_TYPE_EXTSEGADDR = 0x02
HEX_TYPE_STARTSEGADDR = 0x03
HEX_TYPE_EXTLINEARADDR = 0x04
HEX_TYPE_STARTLINEARADDR = 0x05


class SRec(Backend):
    """
    A loader for Motorola SRecord Objects
    See https://en.wikipedia.org/wiki/Intel_HEX
    """

    is_default = True  # Tell CLE to automatically consider using the Hex backend

    @staticmethod
    def calc_checksum(data):
        sum(x.to for x in data)

    @staticmethod
    def parse_record(line):
        addr_size = SREC_ADDR_SIZE[chr(line[1])]
        srec_re = re.compile(srec_regex.format(addr_size=addr_size // 4).encode())
        m = srec_re.match(line)
        if not m:
            raise CLEError(f"Invalid SRec record: {line}")
        my_cksum = 0
        rectype, count, addr, data, cksum = m.groups()
        cksum = int(cksum, 16)
        for d in binascii.unhexlify(line[2:-2]):
            my_cksum = (my_cksum + d) % 256
        my_cksum = 0xFF - my_cksum
        if my_cksum != cksum:
            raise CLEError(f"Invalid checksum: Computed {hex(my_cksum)}, found {hex(cksum)}")
        count = int(count, 16) - ((addr_size // 8) + 1)
        addr = int(addr, 16)
        rectype = int(rectype, 16)
        if data:
            data = binascii.unhexlify(data)
        if data and count != len(data):
            raise CLEError("Data length field does not match length of actual data: " + line)
        return rectype, addr, data

    @staticmethod
    def coalesce_regions(regions):
        # Lots of tiny memory regions is bad!
        # The greedy algorithm to smash them together:
        result = []
        last_addr: int | None = None
        last_data: list[bytes] | None = None
        last_size: int | None = None
        for addr, region in sorted(regions):
            if last_addr is not None and last_addr + last_size == addr:
                last_data.append(region)
                last_size += len(region)
            else:
                if last_addr is not None:
                    result.append((last_addr, b"".join(last_data)))
                last_addr, last_data, last_size = addr, [region], len(region)

        if last_addr is not None:
            result.append((last_addr, b"".join(last_data)))

        return result

    def __init__(self, *args, ignore_missing_arch: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.set_load_args(ignore_missing_arch=ignore_missing_arch)

        if self._arch is None:
            if ignore_missing_arch:
                # used internally for testing. we use a default architecture
                self.set_arch(archinfo.arch_from_id("amd64"))
            else:
                raise CLEError(
                    "To use the Hex binary backend, you need to specify an architecture in the loader options."
                )

        # Do the whole thing in one shot.
        self.os = "unknown"
        got_base = False
        got_entry = False
        self._binary_stream.seek(0)
        string = self._binary_stream.read()
        recs = string.splitlines()
        regions = []
        max_addr = 0
        min_addr = 0xFFFFFFFFFFFFFFFF
        self._base_address = 0
        for rec in recs:
            rectype, addr, data = SRec.parse_record(rec)
            if rectype == SREC_HEADER:
                continue
            if rectype in SREC_DATA:
                addr += self._base_address
                # l.debug("Loading %d bytes at " % len(data) + hex(addr))
                # Raw data.  Put the bytes
                regions.append((addr, data))
                # We have to be careful about the min and max addrs
                min_addr = min(min_addr, addr)
                max_addr = max(max_addr, addr + len(data) - 1)
            elif rectype in SREC_START_EXEC:
                got_entry = True
                self._entry = int.from_bytes(data, "big")
                log.debug("Found entry point at %#x", self._entry)
                self._initial_ip = self._entry
            else:
                raise CLEError("This SRec Object type is not implemented: " + hex(rectype))
        if not got_base:
            log.warning("No base address was found in this SRec object file. It is assumed to be 0")
        if not got_entry:
            log.warning(
                "No entry point was found in this SRec object file, and it is assumed to be 0. "
                "Specify one with `entry_point` to override."
            )
        # HEX specifies a ton of tiny little memory regions.  We now smash them together to make things faster.
        new_regions = SRec.coalesce_regions(regions)
        self.regions: list[tuple[int, int]] = []  # A list of (addr, size)
        for addr, data in new_regions:
            self.memory.add_backer(addr, data)
            self.regions.append((addr, len(data)))
        self._max_addr = max_addr
        self._min_addr = min_addr

    @staticmethod
    def is_compatible(stream):
        stream.seek(0)
        s = stream.read(0x10)
        stream.seek(0)
        return s.startswith(b"S")


register_backend("srec", SRec)



from __future__ import annotations

from .pereloc import PEReloc


class IMAGE_REL_BASED_MIPS_JMPADDR(PEReloc):
    pass


class IMAGE_REL_BASED_MIPS_JMPADDR16(PEReloc):
    pass


relocation_table_mips = {
    5: IMAGE_REL_BASED_MIPS_JMPADDR,
    9: IMAGE_REL_BASED_MIPS_JMPADDR16,
}

__all__ = ("relocation_table_mips",)



"""Relocation types for PPC64.

Reference: http://refspecs.linuxfoundation.org/ELF/ppc64/PPC-elf64abi-1.9.pdf pages 57-59
"""

from __future__ import annotations

import logging

from .elfreloc import ELFReloc
from .generic import (
    GenericAbsoluteAddendReloc,
    GenericIRelativeReloc,
    GenericJumpslotReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

log = logging.getLogger(name=__name__)

# pylint: disable=missing-class-docstring


class R_PPC64_JMP_SLOT(ELFReloc):
    def relocate(self):
        if self.owner.is_ppc64_abiv1:
            # R_PPC64_JMP_SLOT
            # http://osxr.org/glibc/source/sysdeps/powerpc/powerpc64/dl-machine.h?v=glibc-2.15#0405
            # copy an entire function descriptor struct
            addr = self.resolvedby.owner.memory.unpack_word(self.resolvedby.relative_addr)
            toc = self.resolvedby.owner.memory.unpack_word(self.resolvedby.relative_addr + 8)
            aux = self.resolvedby.owner.memory.unpack_word(self.resolvedby.relative_addr + 16)
            self.owner.memory.pack_word(self.relative_addr, addr)
            self.owner.memory.pack_word(self.relative_addr + 8, toc)
            self.owner.memory.pack_word(self.relative_addr + 16, aux)
        else:
            self.owner.memory.pack_word(self.relative_addr, self.resolvedby.rebased_addr)
        return True


class R_PPC64_RELATIVE(GenericRelativeReloc):
    pass


class R_PPC64_IRELATIVE(GenericIRelativeReloc):
    pass


class R_PPC64_ADDR64(GenericAbsoluteAddendReloc):
    pass


class R_PPC64_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_PPC64_DTPMOD64(GenericTLSModIdReloc):
    pass


class R_PPC64_DTPREL64(GenericTLSDoffsetReloc):
    pass


class R_PPC64_TPREL64(GenericTLSOffsetReloc):
    pass


class R_PPC64_REL24(ELFReloc):
    """
    Relocation Type: 10
    Calculation: (S + A - P) >> 2
    Field: low24*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr
        return (S + A - P) >> 2

    def relocate(self):
        if not self.resolved:
            return False
        instr = self.owner.memory.unpack_word(self.relative_addr, size=4) & 0b11111100000000000000000000000011
        imm = self.value & 0xFFFFFF
        self.owner.memory.pack_word(self.relative_addr, instr | (imm << 2), size=4)
        return True


class R_PPC64_TOC16_LO(ELFReloc):
    """
    Relocation Type: 48
    Calculation: #lo(S + A - .TOC.)
    Field: half16
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        if self.owner.ppc64_initial_rtoc is None:
            log.warning(".TOC. value not found")
            return (S + A) & 0xFFFF
        TOC = self.owner.ppc64_initial_rtoc
        return (S + A - TOC) & 0xFFFF

    def relocate(self):
        if not self.resolved:
            return False
        self.owner.memory.pack_word(self.relative_addr, self.value, size=2)
        return True


class R_PPC64_TOC16_HI(ELFReloc):
    """
    Relocation Type: 49
    Calculation: #hi(S + A - .TOC.)
    Field: half16
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        if self.owner.ppc64_initial_rtoc is None:
            log.warning(".TOC. value not found")
            return ((S + A) >> 16) & 0xFFFF
        TOC = self.owner.ppc64_initial_rtoc
        return ((S + A - TOC) >> 16) & 0xFFFF

    def relocate(self):
        if not self.resolved:
            return False
        self.owner.memory.pack_word(self.relative_addr, self.value, size=2)
        return True


class R_PPC64_TOC16_HA(ELFReloc):
    """
    Relocation Type: 50
    Calculation: #ha(S + A - .TOC.)
    Field: half16
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        if self.owner.ppc64_initial_rtoc is None:
            log.warning(".TOC. value not found")
            return (((S + A) >> 16) + (1 if ((S + A) & 0x8000) else 0)) & 0xFFFF
        TOC = self.owner.ppc64_initial_rtoc
        return (((S + A - TOC) >> 16) + (1 if ((S + A - TOC) & 0x8000) else 0)) & 0xFFFF

    def relocate(self):
        if not self.resolved:
            return False
        self.owner.memory.pack_word(self.relative_addr, self.value, size=2)
        return True


class R_PPC64_TOC(ELFReloc):
    """
    Relocation Type: 51
    Calculation: .TOC.
    Field: doubleword64
    """

    @property
    def value(self):
        if self.owner.ppc64_initial_rtoc is None:
            log.warning(".TOC. value not found")
            return 0
        return self.owner.ppc64_initial_rtoc


relocation_table_ppc64 = {
    # 1: R_PPC64_ADDR32,
    # 2: R_PPC64_ADDR24,
    # 3: R_PPC64_ADDR16,
    # 4: R_PPC64_ADDR16_LO,
    # 5: R_PPC64_ADDR16_HI,
    # 6: R_PPC64_ADDR16_HA,
    # 7: R_PPC64_ADDR14,
    # 8: R_PPC64_ADDR14_BRTAKEN,
    # 9: R_PPC64_ADDR14_BRNTAKEN,
    10: R_PPC64_REL24,
    # 11: R_PPC64_REL14,
    # 12: R_PPC64_REL14_BRTAKEN,
    # 13: R_PPC64_REL14_BRNTAKEN,
    # 14: R_PPC64_GOT16,
    # 15: R_PPC64_GOT16_LO,
    # 16: R_PPC64_GOT16_HI,
    # 17: R_PPC64_GOT16_HA,
    # No 18 in doc
    # 19: R_PPC64_COPY,
    20: R_PPC64_GLOB_DAT,
    21: R_PPC64_JMP_SLOT,
    22: R_PPC64_RELATIVE,
    # No 23 in doc
    # 24: R_PPC64_UADDR32,
    # 25: R_PPC64_UADDR16,
    # 26: R_PPC64_REL32,
    # 27: R_PPC64_PLT32,
    # 28: R_PPC64_PLTREL32,
    # 29: R_PPC64_PLT16_LO,
    # 30: R_PPC64_PLT16_HI,
    # 31: R_PPC64_PLT16_HA,
    # No 32 in doc
    # 33: R_PPC64_SECTOFF,
    # 34: R_PPC64_SECTOFF_LO,
    # 35: R_PPC64_SECTOFF_HI,
    # 36: R_PPC64_SECTOFF_HA,
    # 37: R_PPC64_ADDR30,
    38: R_PPC64_ADDR64,
    # 39: R_PPC64_ADDR16_HIGHER,
    # 40: R_PPC64_ADDR16_HIGHERA,
    # 41: R_PPC64_ADDR16_HIGHEST,
    # 42: R_PPC64_ADDR16_HIGHESTA,
    # 43: R_PPC64_UADDR64,
    # 44: R_PPC64_REL64,
    # 45: R_PPC64_PLT64,
    # 46: R_PPC64_PLTREL64,
    # 47: R_PPC64_TOC16,
    48: R_PPC64_TOC16_LO,
    49: R_PPC64_TOC16_HI,
    50: R_PPC64_TOC16_HA,
    51: R_PPC64_TOC,
    # 52: R_PPC64_PLTGOT16,
    # 53: R_PPC64_PLTGOT16_LO,
    # 54: R_PPC64_PLTGOT16_HI,
    # 55: R_PPC64_PLTGOT16_HA,
    # 56: R_PPC64_ADDR16_DS,
    # 57: R_PPC64_ADDR16_LO_DS,
    # 58: R_PPC64_GOT16_DS,
    # 59: R_PPC64_GOT16_LO_DS,
    # 60: R_PPC64_PLT16_LO_DS,
    # 61: R_PPC64_SECTOFF_DS,
    # 62: R_PPC64_SECTOFF_LO_DS,
    # 63: R_PPC64_TOC16_DS,
    # 64: R_PPC64_TOC16_LO_DS,
    # 65: R_PPC64_PLTGOT16_DS,
    # 66: R_PPC64_PLTGOT16_LO_DS,
    # 67: R_PPC64_TLS,
    68: R_PPC64_DTPMOD64,
    # 69: R_PPC64_TPREL16,
    # 70: R_PPC64_TPREL16_LO,
    # 71: R_PPC64_TPREL16_HI,
    # 72: R_PPC64_TPREL16_HA,
    73: R_PPC64_TPREL64,
    # 74: R_PPC64_DTPREL16,
    # 75: R_PPC64_DTPREL16_LO,
    # 76: R_PPC64_DTPREL16_HI,
    # 77: R_PPC64_DTPREL16_HA,
    78: R_PPC64_DTPREL64,
    # 79: R_PPC64_GOT_TLSGD16,
    # 80: R_PPC64_GOT_TLSGD16_LO,
    # 81: R_PPC64_GOT_TLSGD16_HI,
    # 82: R_PPC64_GOT_TLSGD16_HA,
    # 83: R_PPC64_GOT_TLSLD16,
    # 84: R_PPC64_GOT_TLSLD16_LO,
    # 85: R_PPC64_GOT_TLSLD16_HI,
    # 86: R_PPC64_GOT_TLSLD16_HA,
    # 87: R_PPC64_GOT_TPREL16_DS,
    # 88: R_PPC64_GOT_TPREL16_LO_DS,
    # 89: R_PPC64_GOT_TPREL16_HI,
    # 90: R_PPC64_GOT_TPREL16_HA,
    # 91: R_PPC64_GOT_DTPREL16_DS,
    # 92: R_PPC64_GOT_DTPREL16_LO_DS,
    # 93: R_PPC64_GOT_DTPREL16_HI,
    # 94: R_PPC64_GOT_DTPREL16_HA,
    # 95: R_PPC64_TPREL16_DS,
    # 96: R_PPC64_TPREL16_LO_DS,
    # 97: R_PPC64_TPREL16_HIGHER,
    # 98: R_PPC64_TPREL16_HIGHERA,
    # 99: R_PPC64_TPREL16_HIGHEST,
    # 100: R_PPC64_TPREL16_HIGHESTA,
    # 101: R_PPC64_DTPREL16_DS,
    # 102: R_PPC64_DTPREL16_LO_DS,
    # 103: R_PPC64_DTPREL16_HIGHER,
    # 104: R_PPC64_DTPREL16_HIGHERA,
    # 105: R_PPC64_DTPREL16_HIGHEST,
    # 106: R_PPC64_DTPREL16_HIGHESTA,
    # Not in spec
    248: R_PPC64_IRELATIVE,
}

__all__ = ("relocation_table_ppc64",)



from __future__ import annotations

import logging
import struct

from cle.address_translator import AT

from .pereloc import PEReloc

log = logging.getLogger(name=__name__)

__all__ = [
    "DllImport",
    "IMAGE_REL_BASED_ABSOLUTE",
    "IMAGE_REL_BASED_HIGHADJ",
    "IMAGE_REL_BASED_HIGHLOW",
    "IMAGE_REL_BASED_DIR64",
    "IMAGE_REL_BASED_HIGH",
    "IMAGE_REL_BASED_LOW",
]


class DllImport(PEReloc):
    """
    There's nothing special to be done for DLL imports but this class
    provides a unique name to the relocation type.
    """

    pass


class IMAGE_REL_BASED_ABSOLUTE(PEReloc):
    def relocate(self):
        pass


class IMAGE_REL_BASED_HIGHADJ(PEReloc):
    def __init__(self, owner, addr, next_rva):
        super().__init__(owner, None, addr)
        self.next_rva = next_rva

    @property
    def value(self):
        """
        In all the other cases, we can ignore the relocation difference part of the
        calculation because we simply use to_mva() to get our rebased address. In this
        case, however, we have to adjust the un-rebased address first.
        """
        org_bytes = self.owner.memory.load(self.relative_addr, 2)
        org_value = struct.unpack("<H", org_bytes)[0]
        adjusted_value = (org_value << 16) + self.next_rva
        adjusted_value = (AT.from_lva(adjusted_value, self.owner).to_mva() & 0xFFFF0000) >> 16
        adjusted_bytes = struct.pack("<I", adjusted_value)
        return adjusted_bytes


class IMAGE_REL_BASED_HIGHLOW(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 4)
        org_value = struct.unpack("<I", org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        rebased_bytes = struct.pack("<I", rebased_value & 0xFFFFFFFF)
        return rebased_bytes


class IMAGE_REL_BASED_DIR64(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 8)
        org_value = struct.unpack("<Q", org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        if rebased_value < 0 or rebased_value >= 0x10000000000000000:
            log.error(
                "Incorrect rebased address %x found at %s. Maybe the relocation table is broken.",
                rebased_value,
                self.owner,
            )
            return None
        rebased_bytes = struct.pack("<Q", rebased_value)
        return rebased_bytes


class IMAGE_REL_BASED_HIGH(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 2)
        org_value = struct.unpack("<H", org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        adjusted_value = (rebased_value >> 16) & 0xFFFF
        adjusted_bytes = struct.pack("<H", adjusted_value)
        return adjusted_bytes


class IMAGE_REL_BASED_LOW(PEReloc):
    @property
    def value(self):
        org_bytes = self.owner.memory.load(self.relative_addr, 2)
        org_value = struct.unpack("<H", org_bytes)[0]
        rebased_value = AT.from_lva(org_value, self.owner).to_mva()
        adjusted_value = rebased_value & 0x0000FFFF
        adjusted_bytes = struct.pack("<H", adjusted_value)
        return adjusted_bytes


relocation_table_generic = {
    1: IMAGE_REL_BASED_HIGH,
    2: IMAGE_REL_BASED_LOW,
    3: IMAGE_REL_BASED_HIGHLOW,
    4: IMAGE_REL_BASED_HIGHADJ,
    10: IMAGE_REL_BASED_DIR64,
}



"""Relocation types for PowerPC 32-bit architecture.

Reference: http://refspecs.linux-foundation.org/elf/elfspec_ppc.pdf page 4-18

Only relocations 1-37 are described in the document. The rest are from the GNU
binutils source code. See include/elf/ppc.h in the binutils source code.
"""

from __future__ import annotations

import logging

from .elfreloc import ELFReloc
from .generic import (
    GenericAbsoluteAddendReloc,
    GenericCopyReloc,
    GenericJumpslotReloc,
    GenericRelativeReloc,
    GenericTLSDoffsetReloc,
    GenericTLSModIdReloc,
    GenericTLSOffsetReloc,
)

log = logging.getLogger(name=__name__)


# PPC constants/masks to be used in relocations
PPC_WORD32 = 0xFFFFFFFF
PPC_WORD30 = 0xFFFFFFFC
PPC_LOW24 = 0x03FFFFFC
PPC_LOW14 = 0x0020FFFC
PPC_HALF16 = 0xFFFF
PPC_BL_INST = 0x48000001


class R_PPC_ADDR32(GenericAbsoluteAddendReloc):
    pass


class R_PPC_ADDR24(ELFReloc):
    """
    Relocation Type: 0x2
    Calculation: (S + A) >> 2
    Field: low24*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = (S + A) >> 2
        return result


class R_PPC_ADDR16(ELFReloc):
    """
    Relocation Type: 0x3
    Calculation: S+A
    Field: half16*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = S + A
        return result


class R_PPC_ADDR16_LO(ELFReloc):
    """
    Relocation Type: 0x4
    Calculation: #lo(S + A)
    Field: half16
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = S + A
        result = result & PPC_HALF16
        return result

    def relocate(self):
        if not self.resolved:
            return False
        self.owner.memory.pack_word(self.relative_addr, self.value, size=2)
        return True


class R_PPC_ADDR16_HI(ELFReloc):
    """
    Relocation Type: 0x5
    Calculation: #hi(S + A)
    Field: half16
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = (S + A) >> 16
        result = result & PPC_HALF16
        return result

    def relocate(self):
        if not self.resolved:
            return False
        self.owner.memory.pack_word(self.relative_addr, self.value, size=2)
        return True


class R_PPC_ADDR16_HA(ELFReloc):  # pylint: disable=undefined-variable
    """
    Relocation Type: 0x6
    Calculation: #ha(S + A)
    Field: half16
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = S + A
        result = ((result >> 16) + (1 if (result & 0x8000) else 0)) & PPC_HALF16
        return result

    def relocate(self):
        if not self.resolved:
            return False
        self.owner.memory.pack_word(self.relative_addr, self.value, size=2)
        return True


class R_PPC_ADDR14(ELFReloc):
    """
    Relocation Type: 0x7
    Calculation: (S + A) >> 2
    Field: low14*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = (S + A) >> 2
        return result


class R_PPC_ADDR14_BRTAKEN(ELFReloc):
    """
    Relocation Type: 0x8
    Calculation: (S + A) >> 2
    Field: low14*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = (S + A) >> 2
        return result


class R_PPC_ADDR14_BRNTAKEN(ELFReloc):
    """
    Relocation Type: 0x9
    Calculation: (S + A) >> 2
    Field: low14*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = (S + A) >> 2
        return result


class R_PPC_REL24(ELFReloc):
    """
    Relocation Type: 0xa
    Calculation: (S + A - P) >> 2
    Field: low24*
    R_PPC_REL24 is a special type of relocation.
    The instruction must be modified for this type.
    This relocation type resolves branch-and-link instructions.
    Prior to relocation, all instances of the branch-and-link instruction
    will consist of the following bytecode: 48 00 00 01.
    The problem with this is that all instances will result in calls to
    the current address - thus an infinite loop.
    After calculating the relocation result in R_PPC_REL24,
    you will have an address offset to the call.
    The result must be resolved to the correct instruction encoding.
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr

        result = (S + A - P) >> 2
        result = (result << 2) & PPC_LOW24
        result = (A & ~PPC_LOW24) | result
        result = result | PPC_BL_INST
        return result


class R_PPC_REL14(ELFReloc):
    """
    Relocation Type: 0xb
    Calculation: (S + A - P) >> 2
    Field: low14*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr

        result = (S + A - P) >> 2
        result = (result << 2) & PPC_LOW14
        result = (A & ~PPC_LOW14) | result
        return result


class R_PPC_REL14_BRTAKEN(ELFReloc):
    """
    Relocation Type: 0xc
    Calculation: (S + A - P) >> 2
    Field: low14*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr

        result = (S + A - P) >> 2
        result = (result << 2) & PPC_LOW14
        result = (A & ~PPC_LOW14) | result
        return result


class R_PPC_REL14_BRNTAKEN(ELFReloc):
    """
    Relocation Type: 0xd
    Calculation: (S + A - P) >> 2
    Field: low14*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr
        P = self.rebased_addr

        result = (S + A - P) >> 2
        result = (result << 2) & PPC_LOW14
        result = (A & ~PPC_LOW14) | result
        return result


class R_PPC_COPY(GenericCopyReloc):
    pass


class R_PPC_GLOB_DAT(GenericJumpslotReloc):
    pass


class R_PPC_JMP_SLOT(GenericJumpslotReloc):
    def relocate(self):
        if "DT_PPC_GOT" not in self.owner._dynamic and "DT_LOPROC" not in self.owner._dynamic:
            # old PowerPC ABI - we overwrite this location with a jump (b, 0x12. .. .. .1) to the actual target
            val = (0x12 << 26) | ((self.value - self.rebased_addr) & 0x3FFFFFE)
            self.owner.memory.pack_word(self.dest_addr, val)
        else:
            super().relocate()


class R_PPC_RELATIVE(GenericRelativeReloc):
    pass


class R_PPC_UADDR32(ELFReloc):
    """
    Relocation Type: 0x18
    Calculation: S + A
    Field: word32
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = S + A
        return result


class R_PPC_UADDR16(ELFReloc):
    """
    Relocation Type: 0x19
    Calculation: S + A
    Field: half16*
    """

    @property
    def value(self):
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = S + A
        return result


class R_PPC_REL32(ELFReloc):  # pylint: disable=undefined-variable
    """
    Relocation Type: 0x1a
    Calculation: S + A - P
    Field: word32
    """

    @property
    def value(self):
        P = self.rebased_addr
        A = self.addend
        S = self.resolvedby.rebased_addr

        result = (S + A - P) & PPC_WORD32
        return result


class R_PPC_SECTOFF(ELFReloc):
    """
    Relocation Type: 0x21
    Calculation: R + A
    Field: half16*
    """

    @property
    def value(self):
        R = self.relative_addr
        A = self.addend

        result = R + A
        return result


class R_PPC_SECTOFF_LO(ELFReloc):
    """
    Relocation Type: 0x22
    Calculation: #lo(R + A)
    Field: half16
    """

    @property
    def value(self):
        R = self.relative_addr
        A = self.addend

        result = R + A
        result = result & PPC_HALF16
        return result


class R_PPC_SECTOFF_HI(ELFReloc):
    """
    Relocation Type: 0x23
    Calculation: #hi(R + A)
    Field: half16
    """

    @property
    def value(self):
        R = self.relative_addr
        A = self.addend

        result = (R + A) >> 16
        result = result & PPC_HALF16
        return result


class R_PPC_SECTOFF_HA(ELFReloc):
    """
    Relocation Type: 0x24
    Calculation: #ha(R + A)
    Field: half16
    """

    @property
    def value(self):
        R = self.relative_addr
        A = self.addend

        result = R + A
        result = ((result >> 16) + (1 if (result & 0x8000) else 0)) & PPC_HALF16
        return result


class R_PPC_ADDR30(ELFReloc):
    """
    Relocation Type: 0x25
    Calculation: (S + A - P) >> 2
    Field: word30
    """

    @property
    def value(self):
        S = self.resolvedby.rebased_addr
        A = self.addend
        P = self.rebased_addr

        result = (S + A - P) >> 2
        return result


class R_PPC_DTPMOD32(GenericTLSModIdReloc):
    pass


class R_PPC_DTPREL32(GenericTLSDoffsetReloc):
    pass


class R_PPC_TPREL32(GenericTLSOffsetReloc):
    pass


relocation_table_ppc = {
    1: R_PPC_ADDR32,
    2: R_PPC_ADDR24,
    3: R_PPC_ADDR16,
    4: R_PPC_ADDR16_LO,
    5: R_PPC_ADDR16_HI,
    6: R_PPC_ADDR16_HA,
    7: R_PPC_ADDR14,
    8: R_PPC_ADDR14_BRTAKEN,
    9: R_PPC_ADDR14_BRNTAKEN,
    10: R_PPC_REL24,
    11: R_PPC_REL14,
    12: R_PPC_REL14_BRTAKEN,
    13: R_PPC_REL14_BRNTAKEN,
    # 14: R_PPC_GOT16,
    # 15: R_PPC_GOT16_LO,
    # 16: R_PPC_GOT16_HI,
    # 17: R_PPC_GOT16_HA,
    # 18: R_PPC_PLTREL24,
    19: R_PPC_COPY,
    20: R_PPC_GLOB_DAT,
    21: R_PPC_JMP_SLOT,
    22: R_PPC_RELATIVE,
    # 23: R_PPC_LOCAL24PC,
    24: R_PPC_UADDR32,
    25: R_PPC_UADDR16,
    26: R_PPC_REL32,
    # 27: R_PPC_PLT32,
    # 28: R_PPC_PLTREL32,
    # 29: R_PPC_PLT16_LO,
    # 30: R_PPC_PLT16_HI,
    # 31: R_PPC_PLT16_HA,
    # 32: R_PPC_SDAREL16,
    33: R_PPC_SECTOFF,
    34: R_PPC_SECTOFF_LO,
    35: R_PPC_SECTOFF_HI,
    36: R_PPC_SECTOFF_HA,
    37: R_PPC_ADDR30,
    # 67: R_PPC_TLS,
    68: R_PPC_DTPMOD32,
    # 69: R_PPC_TPREL16,
    # 70: R_PPC_TPREL16_LO,
    # 71: R_PPC_TPREL16_HI,
    # 72: R_PPC_TPREL16_HA,
    73: R_PPC_TPREL32,
    # 74: R_PPC_DTPREL16,
    # 75: R_PPC_DTPREL16_LO,
    # 76: R_PPC_DTPREL16_HI,
    # 77: R_PPC_DTPREL16_HA,
    78: R_PPC_DTPREL32,
    # 79: R_PPC_GOT_TLSGD16,
    # 80: R_PPC_GOT_TLSGD16_LO,
    # 81: R_PPC_GOT_TLSGD16_HI,
    # 82: R_PPC_GOT_TLSGD16_HA,
    # 83: R_PPC_GOT_TLSLD16,
    # 84: R_PPC_GOT_TLSLD16_LO,
    # 85: R_PPC_GOT_TLSLD16_HI,
    # 86: R_PPC_GOT_TLSLD16_HA,
    # 87: R_PPC_GOT_TPREL16,
    # 88: R_PPC_GOT_TPREL16_LO,
    # 89: R_PPC_GOT_TPREL16_HI,
    # 90: R_PPC_GOT_TPREL16_HA,
    # 91: R_PPC_GOT_DTPREL16,
    # 92: R_PPC_GOT_DTPREL16_LO,
    # 93: R_PPC_GOT_DTPREL16_HI,
    # 94: R_PPC_GOT_DTPREL16_HA,
    # 95: R_PPC_TLSGD,
    # 96: R_PPC_TLSLD,
}

__all__ = ("relocation_table_ppc",)



from __future__ import annotations

import logging

from cle.backends.relocation import Relocation

log = logging.getLogger(name=__name__)


# Reference: https://msdn.microsoft.com/en-us/library/ms809762.aspx
class PEReloc(Relocation):
    AUTO_HANDLE_NONE = True

    def __init__(self, owner, symbol, addr, resolvewith=None):  # pylint: disable=unused-argument
        super().__init__(owner, symbol, addr)

        self.resolvewith = resolvewith
        if self.resolvewith is not None:
            self.resolvewith = self.resolvewith.lower()

    def resolve_symbol(self, solist, bypass_compatibility=False, extern_object=None, **kwargs):
        if not bypass_compatibility:
            solist = [x for x in solist if self.resolvewith == x.provides]
        super().resolve_symbol(solist, bypass_compatibility=bypass_compatibility, extern_object=extern_object, **kwargs)

        if self.resolvedby is None:
            return

        # handle symbol forwarders
        newsym = self.resolvedby.resolve_forwarder()
        if newsym is None:
            new_symbol = extern_object.make_extern(self.symbol.name, sym_type=self.symbol.type)
            self.resolvedby.resolvedby = new_symbol
            self.resolve(new_symbol)
            return

        self.resolvedby = newsym
        self.symbol.resolvedby = newsym

    def relocate(self):
        if self.symbol is None:  # relocation described in the DIRECTORY_ENTRY_BASERELOC table
            value = self.value
            if value is None:
                log.debug("Unresolved relocation with no symbol.")
                return
            self.owner.memory.store(self.relative_addr, value)
        else:
            super().relocate()

    @property
    def value(self):
        if self.resolved:
            return self.resolvedby.rebased_addr

    @property
    def is_base_reloc(self):
        """
        These relocations are ignored by the linker if the executable
        is loaded at its preferred base address. There is no associated
        symbol with base relocations.
        """
        return True if self.symbol is None else False

    @property
    def is_import(self):
        return not self.is_base_reloc



from __future__ import annotations

import struct

from cle.backends.relocation import Relocation
from cle.backends.symbol import SymbolType

from .simdata import SimData


class StaticData(SimData):
    """
    A simple SimData utility class to use when you have a SimData which should provide just a static
    set of bytes. To use, implement the following:

    :cvar name:     The name of the symbol to provide.
    :cvar libname:  The name of the library from which the symbol originally comes (currently unused).
    :cvar data:     The bytes to provide
    """

    type = SymbolType.TYPE_OBJECT
    data: bytes = NotImplemented

    @classmethod
    def static_size(cls, owner):
        return len(cls.data)

    def value(self):
        return self.data


class StaticWord(SimData):
    """
    A simple SimData utility class to use when you have a SimData which should provide just a static
    integer. To use, implement the following:

    :cvar name:     The name of the symbol to provide.
    :cvar libname:  The name of the library from which the symbol originally comes (currently unused).
    :cvar word:     The value to provide
    :cvar wordsize: (optional) The size of the value in bytes, default the CPU wordsize
    """

    type = SymbolType.TYPE_OBJECT
    word: int = NotImplemented
    wordsize: int = None

    @classmethod
    def static_size(cls, owner):
        return owner.arch.bytes if cls.wordsize is None else cls.wordsize

    def value(self):
        return struct.pack(self.owner.arch.struct_fmt(size=self.wordsize), self.word)


class PointTo(SimData):
    """
    A simple SimData utility class to use when you have a SimData which should provide just a
    pointer to some other symbol. To use, implement the following:

    :cvar name:         The name of the symbol to provide.
    :cvar libname:      The name of the library from which the symbol originally comes (currently unused).
    :cvar pointto_name: The name of the symbol to point to
    :cvar pointto_type: The type of the symbol to point to (usually ``SymbolType.TYPE_FUNCTION`` or
                        ``SymbolType.TYPE_OBJECT``)
    :cvar addend:       (optional) an integer to be added to the symbol's address before storage
    """

    pointto_name: str = NotImplemented
    pointto_type: SymbolType = NotImplemented
    type: SymbolType = SymbolType.TYPE_OBJECT
    addend: int = 0

    @classmethod
    def static_size(cls, owner):
        return owner.arch.bytes

    def value(self):
        return bytes(self.size)

    def relocations(self):
        return [
            SimDataSimpleRelocation(
                self.owner,
                self.owner.make_import(self.pointto_name, self.pointto_type),
                self.relative_addr,
                self.addend,
            )
        ]


class SimDataSimpleRelocation(Relocation):
    """
    A relocation used to implement PointTo. Pretty simple.
    """

    def __init__(self, owner, symbol, addr, addend, preresolved=False):
        super().__init__(owner, symbol, addr)
        self.addend = addend
        self.preresolved = preresolved

    def resolve_symbol(self, solist, **kwargs):
        if self.preresolved:
            self.resolve(self.symbol)
        else:
            super().resolve_symbol(solist, **kwargs)

    @property
    def value(self):
        return self.resolvedby.rebased_addr + self.addend



from __future__ import annotations

from .pereloc import PEReloc


class IMAGE_REL_BASED_RISCV_HIGH20(PEReloc):
    pass


class IMAGE_REL_BASED_RISCV_LOW12I(PEReloc):
    pass


class IMAGE_REL_BASED_RISCV_LOW12S(PEReloc):
    pass


relocation_table_riscv = {
    5: IMAGE_REL_BASED_RISCV_HIGH20,
    7: IMAGE_REL_BASED_RISCV_LOW12I,
    8: IMAGE_REL_BASED_RISCV_LOW12S,
}

__all__ = ("relocation_table_riscv",)



"""Relocations for SPARC

Reference: https://sparc.org/wp-content/uploads/2014/01/psABI3rd.pdf.gz page 4-4
"""

from __future__ import annotations

from .elfreloc import ELFReloc


class R_SPARC_HI22(ELFReloc):
    """
    Value: 9
    Field: T-imm22
    Calculation: (S + A) >> 10
    """

    @property
    def value(self):
        result = (self.resolvedby.rebased_addr + self.addend) >> 10
        instr_bytes = self.owner.memory.load(self.relative_addr, 4)
        instr = int.from_bytes(instr_bytes, byteorder="big")
        return instr & 0xFFC00000 | result & 0x3FFFFF


class R_SPARC_WDISP30(ELFReloc):
    """
    Value: 7
    Field: V-disp30
    Calculation: (S + A - P) >> 2
    """

    @property
    def value(self):
        result = (self.resolvedby.rebased_addr + self.addend - self.rebased_addr) >> 2
        instr_bytes = self.owner.memory.load(self.relative_addr, 4)
        instr = int.from_bytes(instr_bytes, byteorder="big")
        return instr & 0xC0000000 | result & 0x3FFFFFFF


class R_SPARC_LO10(ELFReloc):
    """
    Value: 12
    Field: T-simm13
    Calculation: (S + A) & 0x3ff
    """

    @property
    def value(self):
        result = (self.resolvedby.rebased_addr + self.addend) & 0x3FF
        instr_bytes = self.owner.memory.load(self.relative_addr, 4)
        instr = int.from_bytes(instr_bytes, byteorder="big")
        return instr & 0xFFFFE000 | result & 0x1FFF


relocation_table_sparc = {
    # 1: R_SPARC_8,
    # 2: R_SPARC_16,
    # 3: R_SPARC_32,
    # 4: R_SPARC_DISP8,
    # 5: R_SPARC_DISP16,
    # 6: R_SPARC_DISP32,
    7: R_SPARC_WDISP30,
    # 8: R_SPARC_WDISP22,
    9: R_SPARC_HI22,
    # 10: R_SPARC_22,
    # 11: R_SPARC_13,
    12: R_SPARC_LO10,
    # 13: R_SPARC_GOT10,
    # 14: R_SPARC_GOT13,
    # 15: R_SPARC_GOT22,
    # 16: R_SPARC_PC10,
    # 17: R_SPARC_PC22,
    # 18: R_SPARC_WPLT30,
    # 19: R_SPARC_COPY,
    # 20: R_SPARC_GLOB_DAT,
    # 21: R_SPARC_JMP_SLOT,
    # 22: R_SPARC_RELATIVE,
    # 23: R_SPARC_UA32,
}

__all__ = ("relocation_table_sparc",)



from __future__ import annotations

from . import glibc_startup, io_file
from .simdata import SimData, lookup, register

__all__ = [
    "SimData",
    "lookup",
    "register",
    "glibc_startup",
    "io_file",
]
